# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import hashlib
import os
from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from functools import cache
from pathlib import Path
from typing import Any

_LOCAL_CACHE_ENV_VAR = "SCIPP_DATA_DIR"
_LOCAL_REGISTRY_ENV_VAR = "SCIPP_OVERRIDE_DATA_DIR"


def make_registry(
    prefix: str,
    files: Mapping[str, str | Entry],
    *,
    version: str,
    base_url: str = "https://public.esss.dk/groups/scipp",
    retry_if_failed: int = 3,
) -> Registry:
    """Create a file registry object.

    By default, this function creates a :class:`PoochRegistry` to download files
    via HTTP from an online file store.
    This can be overridden by setting the environment variable `SCIPP_DATA_DIR` to a
    path on the local file system.
    In this case, a :class:`LocalRegistry` is returned.

    Files are specified as a dict using either the Pooch string format explicitly
    constructed :class:`Entry` objects:

        >>> from ess.reduce.data import Entry
        >>> files = {
        ...    "file1.dat": "md5:1234567890abcdef",
        ...    "file2.csv": Entry(alg="md5", chk="abcdef123456789"),
        ...    "folder/nested.dat": "blake2b:1234567890abcdef",
        ...    "zipped.zip": Entry(alg="blake2b", chk="abcdef123456789", unzip=True),
        ... }

    In the example above, the specifications for ``file1.dat`` and ``file2.csv`` are
    essentially equivalent.
    ``folder/nested.dat`` is a file in a subfolder.
    Paths like this must always use forward slashes (/) even on Windows.

    As shown above, it is possible to automatically unzip
    files by specifying ``unzip=True``.
    When calling ``registry.get_path("zipped.zip")`` the file will be unzipped and
    a path to the content is returned.
    This expects that there is only a single file in the zip archive.

    The complete path to the source file is constructed as follows:

    - Pooch: ``{base_url}/{prefix}/{version}/{name}``
    - Local: ``{SCIPP_OVERRIDE_DATA_DIR}/{prefix}/{version}/{name}``

    When using Pooch, files are downloaded to the user's cache directory.
    This can be controlled with the environment variable ``SCIPP_CACHE_DIR``.

    Parameters
    ----------
    prefix:
        Prefix to add to all file names.
    files:
        Mapping of file names to checksums or :class:`Entry` objects.
    version:
        A version string for the files.
    base_url:
        URL for the online file store.
        Ignored if the override environment variable is set.
    retry_if_failed:
        Number of retries when downloading a file.
        Ignored if the override environment variable is set.

    Returns
    -------
    :
        Either a :class:`PoochRegistry` or :class:`LocalRegistry`.
    """
    if (override := os.environ.get(_LOCAL_REGISTRY_ENV_VAR)) is not None:
        return LocalRegistry(
            _check_local_override_path(override),
            prefix,
            files,
            version=version,
            base_url=base_url,
            retry_if_failed=retry_if_failed,
        )
    return PoochRegistry(
        prefix,
        files,
        version=version,
        base_url=base_url,
        retry_if_failed=retry_if_failed,
    )


def _check_local_override_path(override: str) -> Path:
    path = Path(override)
    if not path.is_dir():
        raise ValueError(
            f"The data override path '{override}' is not a directory. If you want to "
            "download files instead, unset the environment variable "
            f"{_LOCAL_REGISTRY_ENV_VAR}."
        )
    return path


@dataclass(frozen=True, slots=True)
class Entry:
    """An entry in a registry."""

    chk: str
    """Checksum."""
    alg: str
    """Checksum algorithm."""
    unzip: bool = False
    """Whether to unzip the file."""

    @classmethod
    def from_pooch_string(cls, pooch_string: str) -> Entry:
        alg, chk = pooch_string.split(":")
        return cls(chk=chk, alg=alg)


class Registry(ABC):
    def __init__(self, files: Mapping[str, str | Entry]) -> None:
        self._files = _to_file_entries(files)

    @abstractmethod
    def get_path(self, name: str) -> Path:
        """Get the path to a file in the registry.

        Depending on the implementation, the file is downloaded if necessary.

        Note that implementations are allowed to cache return values of this method
        to avoid recomputing potentially expensive checksums.
        This usually means that the ``Registry`` object itself gets stored until the
        Python interpreter shuts down.
        However, registries are small and do not own resources.

        Parameters
        ----------
        name:
            Name of the file to get the path for.

        Returns
        -------
        :
            The Path to the file.
        """

    def _needs_unzip(self, name: str) -> bool:
        return self._files[name].unzip


class PoochRegistry(Registry):
    def __init__(
        self,
        prefix: str,
        files: Mapping[str, str | Entry],
        *,
        version: str,
        base_url: str,
        retry_if_failed: int = 3,
    ) -> None:
        self._registry = _create_pooch(
            prefix,
            files,
            version=version,
            base_url=base_url,
            retry_if_failed=retry_if_failed,
        )
        self._unzip_processor = _import_pooch().Unzip()
        super().__init__(files)

    @cache  # noqa: B019
    def get_path(self, name: str) -> Path:
        """Get the path to a file in the registry.

        Downloads the file if necessary.
        """
        if self._needs_unzip(name):
            paths: list[str] = self._registry.fetch(  # type: ignore[assignment]
                name, processor=self._unzip_processor
            )
            return Path(_expect_single_unzipped(paths, name))
        return Path(self._registry.fetch(name))


class LocalRegistry(Registry):
    def __init__(
        self,
        source_path: Path,
        prefix: str,
        files: Mapping[str, str | Entry],
        *,
        version: str,
        base_url: str,
        retry_if_failed: int = 3,
    ) -> None:
        # Piggyback off of Pooch to determine the cache directory.
        pooch_registry = _create_pooch(
            prefix,
            files,
            version=version,
            base_url=base_url,
            retry_if_failed=retry_if_failed,
        )
        pooch = _import_pooch()
        self._unzip_processor = pooch.processors.Unzip(extract_dir=pooch_registry.path)
        self._source_path = source_path.resolve().joinpath(*prefix.split("/"), version)
        super().__init__(files)

    @cache  # noqa: B019
    def get_path(self, name: str) -> Path:
        """Get the path to a file in the registry."""
        try:
            entry = self._files[name]
        except KeyError:
            raise ValueError(f"File '{name}' is not in the registry.") from None

        path = self._local_path(name)
        if not path.exists():
            raise FileNotFoundError(
                f"File '{name}' is registered but does not exist on the file system. "
                f"Expected it at '{path}'."
            )

        _check_hash(name, path, entry)

        if self._needs_unzip(name):
            return Path(
                _expect_single_unzipped(
                    self._unzip_processor(os.fspath(path), "download", None), path
                )
            )
        return path

    def _local_path(self, name: str) -> Path:
        # Split on "/" because `name` is always a POSIX-style path, but the return
        # value is a system path, i.e., it can be a Windows-style path.
        return self._source_path.joinpath(*name.split("/"))


def _import_pooch() -> Any:
    try:
        import pooch
    except ImportError:
        raise ImportError(
            "You need to install Pooch to access test and tutorial files. "
            "See https://www.fatiando.org/pooch/latest/index.html"
        ) from None

    return pooch


def _create_pooch(
    prefix: str,
    files: Mapping[str, str | Entry],
    *,
    version: str,
    base_url: str,
    retry_if_failed: int = 3,
) -> Any:
    pooch = _import_pooch()
    return pooch.create(
        path=pooch.os_cache(prefix),
        env=_LOCAL_CACHE_ENV_VAR,
        base_url=f'{base_url}/{prefix}/{version}/',
        registry=_to_pooch_registry(files),
        retry_if_failed=retry_if_failed,
    )


def _pooch_unzip_processor(extract_dir: Path) -> Any:
    try:
        import pooch
    except ImportError:
        raise ImportError("You need to install Pooch to unzip files.") from None

    return pooch.processors.Unzip(extract_dir=os.fspath(extract_dir))


def _expect_single_unzipped(paths: list[str], archive: str | os.PathLike) -> str:
    if len(paths) != 1:
        raise ValueError(
            f"Expected exactly one file to unzip, got {len(paths)} in "
            f"'{os.fspath(archive)}'."
        )
    return paths[0]


def _check_hash(name: str, path: Path, entry: Entry) -> None:
    new_chk = _checksum_of_file(path, algorithm=entry.alg)
    if new_chk.lower() != entry.chk.lower():
        raise ValueError(
            f"{entry.alg} hash of file '{name}' does not match the known hash: "
            f"expected {entry.chk}, got {new_chk}."
        )


def _to_file_entries(files: Mapping[str, str | Entry]) -> dict[str, Entry]:
    return {
        name: entry if isinstance(entry, Entry) else Entry.from_pooch_string(entry)
        for name, entry in files.items()
    }


def _to_pooch_registry(files: Mapping[str, str | Entry]) -> dict[str, str]:
    return {
        name: f"{entry.alg}:{entry.chk}" if isinstance(entry, Entry) else entry
        for name, entry in files.items()
    }


# Code taken from Scitacean and Pooch.
def _checksum_of_file(path: Path, *, algorithm: str) -> str:
    """Compute the checksum of a local file.

    Parameters
    ----------
    path:
        Path of the file.
    algorithm:
        Hash algorithm to use.
        Can be any algorithm supported by :func:`hashlib.new`.

    Returns
    -------
    :
        The hex digest of the hash.
    """
    chk = _new_hash(algorithm)
    # size based on http://git.savannah.gnu.org/gitweb/?p=coreutils.git;a=blob;f=src/ioblksize.h;h=ed2f4a9c4d77462f357353eb73ee4306c28b37f1;hb=HEAD#l23  # noqa: E501
    buffer = memoryview(bytearray(128 * 1024))
    with open(path, "rb", buffering=0) as file:
        for n in iter(lambda: file.readinto(buffer), 0):
            chk.update(buffer[:n])
    return chk.hexdigest()  # type: ignore[no-any-return]


def _new_hash(algorithm: str) -> Any:
    # Try to use a named constructor instead of hashlib.new where possible
    # because that is supposed to be faster, according to
    # https://docs.python.org/3/library/hashlib.html#hashlib.new
    try:
        return getattr(hashlib, algorithm)()
    except AttributeError:
        return hashlib.new(algorithm, usedforsecurity=False)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import numpy as np

from ess.reduce.time_of_flight.interpolator_numba import (
    Interpolator as InterpolatorNumba,
)
from ess.reduce.time_of_flight.interpolator_scipy import (
    Interpolator as InterpolatorScipy,
)


def _f(x, y):
    """
    Function to interpolate, copied from Scipy docs. See
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.RegularGridInterpolator.html
    """

    return 2 * x**3 + 3 * y**2


def _make_interpolators():
    time_edges = np.linspace(0, 71, 101)
    distance_edges = np.linspace(40, 70, 201)
    time_g, distance_g = np.meshgrid(
        time_edges, distance_edges, indexing='ij', sparse=True
    )
    values = _f(time_g, distance_g).T

    numba_interp = InterpolatorNumba(
        time_edges=time_edges, distance_edges=distance_edges, values=values
    )

    scipy_interp = InterpolatorScipy(
        time_edges=time_edges, distance_edges=distance_edges, values=values
    )
    return numba_interp, scipy_interp


def test_numba_and_scipy_interpolators_yield_same_results():
    numba_interp, scipy_interp = _make_interpolators()

    rng = np.random.default_rng(seed=42)
    npoints = 1000
    times = rng.uniform(0, 71, npoints)
    distances = rng.uniform(40, 70, npoints)

    numba_result = numba_interp(times, distances)
    scipy_result = scipy_interp(times, distances)

    # Do not use equal_nan because there should be no NaNs here
    assert np.allclose(numba_result, scipy_result)


def test_numba_and_scipy_interpolators_yield_same_results_with_pulse_offset():
    numba_interp, scipy_interp = _make_interpolators()

    rng = np.random.default_rng(seed=42)
    npoints = 1000
    times = rng.uniform(0, 71, npoints)
    distances = rng.uniform(40, 70, npoints)
    offsets = rng.uniform(0, 2, npoints)
    period = 1.0

    numba_result = numba_interp(times, distances, period, offsets)
    scipy_result = scipy_interp(times, distances, period, offsets)

    assert np.allclose(numba_result, scipy_result, equal_nan=True)


def test_numba_and_scipy_interpolators_yield_same_results_with_out_of_bounds():
    numba_interp, scipy_interp = _make_interpolators()

    rng = np.random.default_rng(seed=42)
    npoints = 1000
    times = rng.uniform(-1, 72, npoints)
    distances = rng.uniform(39, 71, npoints)

    numba_result = numba_interp(times, distances)
    scipy_result = scipy_interp(times, distances)

    assert np.allclose(numba_result, scipy_result, equal_nan=True)


def test_numba_and_scipy_interpolators_yield_same_results_with_values_on_edges():
    numba_interp, scipy_interp = _make_interpolators()

    rng = np.random.default_rng(seed=42)
    npoints = 2

    times = np.array([0.0, 71.0])
    distances = rng.uniform(39, 71, npoints)
    numba_result = numba_interp(times, distances)
    scipy_result = scipy_interp(times, distances)
    assert np.allclose(numba_result, scipy_result, equal_nan=True)

    times = rng.uniform(0, 71, npoints)
    distances = np.array([40.0, 70.0])
    numba_result = numba_interp(times, distances)
    scipy_result = scipy_interp(times, distances)
    assert np.allclose(numba_result, scipy_result, equal_nan=True)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import numpy as np
from numba import njit, prange


@njit(boundscheck=False, cache=True, fastmath=False, parallel=True)
def interpolate(
    x: np.ndarray,
    y: np.ndarray,
    values: np.ndarray,
    xp: np.ndarray,
    yp: np.ndarray,
    xoffset: np.ndarray | None,
    deltax: float,
    fill_value: float,
    out: np.ndarray,
):
    """
    Linear interpolation of data on a 2D regular grid.

    Parameters
    ----------
    x:
        1D array of grid edges along the x-axis (size nx). They must be linspaced.
    y:
        1D array of grid edges along the y-axis (size ny). They must be linspaced.
    values:
        2D array of values on the grid. The shape must be (ny, nx).
    xp:
        1D array of x-coordinates where to interpolate (size N).
    yp:
        1D array of y-coordinates where to interpolate (size N).
    xoffset:
        1D array of integer offsets to apply to the x-coordinates (size N).
    deltax:
        Multiplier to apply to the integer offsets (i.e. the step size).
    fill_value:
        Value to use for points outside of the grid.
    out:
        1D array where the interpolated values will be stored (size N).
    """
    if not (len(xp) == len(yp) == len(out)):
        raise ValueError("Interpolator: all input arrays must have the same size.")

    nx = len(x)
    ny = len(y)
    npoints = len(xp)
    xmin = x[0]
    xmax = x[nx - 1]
    ymin = y[0]
    ymax = y[ny - 1]
    dx = x[1] - xmin
    dy = y[1] - ymin

    one_over_dx = 1.0 / dx
    one_over_dy = 1.0 / dy
    norm = one_over_dx * one_over_dy

    for i in prange(npoints):
        xx = xp[i] + (xoffset[i] * deltax if xoffset is not None else 0.0)
        yy = yp[i]

        if (xx < xmin) or (xx > xmax) or (yy < ymin) or (yy > ymax):
            out[i] = fill_value

        else:
            ix = nx - 2 if xx == xmax else int((xx - xmin) * one_over_dx)
            iy = ny - 2 if yy == ymax else int((yy - ymin) * one_over_dy)

            x1 = x[ix]
            x2 = x[ix + 1]
            y1 = y[iy]
            y2 = y[iy + 1]

            a11 = values[iy, ix]
            a21 = values[iy, ix + 1]
            a12 = values[iy + 1, ix]
            a22 = values[iy + 1, ix + 1]

            x2mxx = x2 - xx
            xxmx1 = xx - x1

            out[i] = (
                (y2 - yy) * (x2mxx * a11 + xxmx1 * a21)
                + (yy - y1) * (x2mxx * a12 + xxmx1 * a22)
            ) * norm


class Interpolator:
    def __init__(
        self,
        time_edges: np.ndarray,
        distance_edges: np.ndarray,
        values: np.ndarray,
        fill_value: float = np.nan,
    ):
        """
        Interpolator for 2D regular grid data (Numba implementation).

        Parameters
        ----------
        time_edges:
            1D array of time edges.
        distance_edges:
            1D array of distance edges.
        values:
            2D array of values on the grid. The shape must be (ny, nx).
        fill_value:
            Value to use for points outside of the grid.
        """
        self.time_edges = time_edges
        self.distance_edges = distance_edges
        self.values = values
        self.fill_value = fill_value

    def __call__(
        self,
        times: np.ndarray,
        distances: np.ndarray,
        pulse_period: float = 0.0,
        pulse_index: np.ndarray | None = None,
    ) -> np.ndarray:
        out = np.empty_like(times)
        interpolate(
            x=self.time_edges,
            y=self.distance_edges,
            values=self.values,
            xp=times,
            yp=distances,
            xoffset=pulse_index,
            deltax=pulse_period,
            fill_value=self.fill_value,
            out=out,
        )
        return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""
Utilities for computing real neutron time-of-flight from chopper settings and
neutron time-of-arrival at the detectors.
"""

from ..nexus.types import DiskChoppers
from .eto_to_tof import providers
from .lut import (
    DistanceResolution,
    LookupTableRelativeErrorThreshold,
    LtotalRange,
    NumberOfSimulatedNeutrons,
    PulsePeriod,
    PulseStride,
    SimulationResults,
    SimulationSeed,
    SourcePosition,
    TimeResolution,
    TofLookupTableWorkflow,
    simulate_chopper_cascade_using_tof,
)
from .types import (
    DetectorLtotal,
    DetectorTofData,
    MonitorLtotal,
    MonitorTofData,
    PulseStrideOffset,
    TimeOfFlightLookupTable,
    TimeOfFlightLookupTableFilename,
)
from .workflow import GenericTofWorkflow

__all__ = [
    "DetectorLtotal",
    "DetectorTofData",
    "DiskChoppers",
    "DistanceResolution",
    "GenericTofWorkflow",
    "LookupTableRelativeErrorThreshold",
    "LtotalRange",
    "MonitorLtotal",
    "MonitorTofData",
    "NumberOfSimulatedNeutrons",
    "PulsePeriod",
    "PulseStride",
    "PulseStrideOffset",
    "SimulationResults",
    "SimulationSeed",
    "SourcePosition",
    "TimeOfFlightLookupTable",
    "TimeOfFlightLookupTableFilename",
    "TimeResolution",
    "TofLookupTableWorkflow",
    "providers",
    "simulate_chopper_cascade_using_tof",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from ._string_widget import MultiStringWidget, StringWidget


class FilenameWidget(StringWidget): ...


class MultiFilenameWidget(MultiStringWidget): ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from pathlib import Path

import pytest

from ess.reduce.data import Registry, make_registry


@pytest.fixture(scope='session')
def bifrost_registry() -> Registry:
    return make_registry(
        'ess/bifrost',
        files={
            "BIFROST_20240914T053723.h5": "md5:0f2fa5c9a851f8e3a4fa61defaa3752e",
        },
        version='1',
    )


@pytest.fixture(scope='session')
def dream_registry() -> Registry:
    return make_registry(
        'ess/dream',
        files={
            "TEST_977695_00068064.hdf": "md5:9e6ee9ec70d7c5e8c0c93b9e07e8949f",
        },
        version='2',
    )


@pytest.fixture(scope='session')
def loki_registry() -> Registry:
    return make_registry(
        'ess/loki',
        files={
            # Files from LoKI@Larmor detector test experiment
            #
            # Background run 1 (no sample, sample holder/can only, no transmission monitor)  # noqa: E501
            '60248-2022-02-28_2215.nxs': 'md5:d9f17b95274a0fc6468df7e39df5bf03',
            # Sample run 1 (sample + sample holder/can, no transmission monitor in beam)
            '60250-2022-02-28_2215.nxs': 'md5:6a519ceaacbae702a6d08241e86799b1',
            # Sample run 2 (sample + sample holder/can, no transmission monitor in beam)
            '60339-2022-02-28_2215.nxs': 'md5:03c86f6389566326bb0cbbd80b8f8c4f',
            # Background transmission run (sample holder/can + transmission monitor)
            '60392-2022-02-28_2215.nxs': 'md5:9ecc1a9a2c05a880144afb299fc11042',
            # Background run 2 (no sample, sample holder/can only, no transmission monitor)  # noqa: E501
            '60393-2022-02-28_2215.nxs': 'md5:bf550d0ba29931f11b7450144f658652',
            # Sample transmission run (sample + sample holder/can + transmission monitor)  # noqa: E501
            '60394-2022-02-28_2215.nxs': 'md5:c40f38a62337d86957af925296c4c615',
            # Analytical model for the I(Q) of the Poly-Gauss sample
            'PolyGauss_I0-50_Rg-60.h5': 'md5:f5d60d9c2286cb197b8cd4dc82db3d7e',
            # XML file for the pixel mask
            'mask_new_July2022.xml': 'md5:421b6dc9db74126ffbc5d88164d017b0',
        },
        version='2',
    )


@pytest.fixture(scope='session')
def bifrost_simulated_elastic(bifrost_registry: Registry) -> Path:
    """McStas simulation with elastic incoherent scattering + phonon."""
    return bifrost_registry.get_path('BIFROST_20240914T053723.h5')


@pytest.fixture(scope='session')
def loki_tutorial_sample_run_60250(loki_registry: Registry) -> Path:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return loki_registry.get_path('60250-2022-02-28_2215.nxs')


@pytest.fixture(scope='session')
def loki_tutorial_sample_run_60339(loki_registry: Registry) -> Path:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return loki_registry.get_path('60339-2022-02-28_2215.nxs')


@pytest.fixture(scope='session')
def loki_tutorial_background_run_60248(loki_registry: Registry) -> Path:
    """Background run with sample holder/can only, no transmission monitor."""
    return loki_registry.get_path('60248-2022-02-28_2215.nxs')


@pytest.fixture(scope='session')
def loki_tutorial_background_run_60393(loki_registry: Registry) -> Path:
    """Background run with sample holder/can only, no transmission monitor."""
    return loki_registry.get_path('60393-2022-02-28_2215.nxs')


@pytest.fixture(scope='session')
def loki_tutorial_sample_transmission_run(loki_registry: Registry) -> Path:
    """Sample transmission run (sample + sample holder/can + transmission monitor)."""
    return loki_registry.get_path('60394-2022-02-28_2215.nxs')


@pytest.fixture(scope='session')
def dream_coda_test_file(dream_registry: Registry) -> Path:
    """CODA file for DREAM where most pulses have been removed.

    See ``tools/shrink_nexus.py``.
    """
    return dream_registry.get_path('TEST_977695_00068064.hdf')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
Time-of-flight workflow for unwrapping the time of arrival of the neutron at the
detector.
This workflow is used to convert raw detector data with event_time_zero and
event_time_offset coordinates to data with a time-of-flight coordinate.
"""

from collections.abc import Callable

import numpy as np
import scipp as sc
import scippneutron as scn
from scippneutron._utils import elem_unit

try:
    from .interpolator_numba import Interpolator as InterpolatorImpl
except ImportError:
    from .interpolator_scipy import Interpolator as InterpolatorImpl

from ..nexus.types import (
    CalibratedBeamline,
    CalibratedMonitor,
    DetectorData,
    MonitorData,
    MonitorType,
    RunType,
)
from .resample import rebin_strictly_increasing
from .types import (
    DetectorLtotal,
    DetectorTofData,
    MonitorLtotal,
    MonitorTofData,
    PulseStrideOffset,
    TimeOfFlightLookupTable,
)


class TofInterpolator:
    def __init__(self, lookup: sc.DataArray, distance_unit: str, time_unit: str):
        self._distance_unit = distance_unit
        self._time_unit = time_unit

        self._time_edges = (
            lookup.coords["event_time_offset"]
            .to(unit=self._time_unit, copy=False)
            .values
        )
        self._distance_edges = (
            lookup.coords["distance"].to(unit=distance_unit, copy=False).values
        )

        self._interpolator = InterpolatorImpl(
            time_edges=self._time_edges,
            distance_edges=self._distance_edges,
            values=lookup.data.to(unit=self._time_unit, copy=False).values,
        )

    def __call__(
        self,
        ltotal: sc.Variable,
        event_time_offset: sc.Variable,
        pulse_period: sc.Variable,
        pulse_index: sc.Variable | None = None,
    ) -> sc.Variable:
        if ltotal.unit != self._distance_unit:
            raise sc.UnitError(
                f"ltotal must have unit: {self._distance_unit}, "
                f"but got unit: {ltotal.unit}."
            )
        if event_time_offset.unit != self._time_unit:
            raise sc.UnitError(
                f"event_time_offset must have unit: {self._time_unit}, "
                f"but got unit: {event_time_offset.unit}."
            )
        out_dims = event_time_offset.dims
        ltotal = ltotal.values
        event_time_offset = event_time_offset.values

        return sc.array(
            dims=out_dims,
            values=self._interpolator(
                times=event_time_offset,
                distances=ltotal,
                pulse_index=pulse_index.values if pulse_index is not None else None,
                pulse_period=pulse_period.value,
            ),
            unit=self._time_unit,
        )


def _time_of_flight_data_histogram(
    da: sc.DataArray, lookup: sc.DataArray, ltotal: sc.Variable
) -> sc.DataArray:
    # In NeXus, 'time_of_flight' is the canonical name in NXmonitor, but in some files,
    # it may be called 'tof' or 'frame_time'.
    key = next(iter(set(da.coords.keys()) & {"time_of_flight", "tof", "frame_time"}))
    raw_eto = da.coords[key].to(dtype=float, copy=False)
    eto_unit = raw_eto.unit
    pulse_period = lookup.coords["pulse_period"].to(unit=eto_unit)

    # In histogram mode, because there is a wrap around at the end of the pulse, we
    # need to insert a bin edge at that exact location to avoid having the last bin
    # with one finite left edge and a NaN right edge (it becomes NaN as it would be
    # outside the range of the lookup table).
    new_bins = sc.sort(
        sc.concat([raw_eto, sc.scalar(0.0, unit=eto_unit), pulse_period], dim=key),
        key=key,
    )
    rebinned = da.rebin({key: new_bins})
    etos = rebinned.coords[key]

    # Create linear interpolator
    interp = TofInterpolator(lookup, distance_unit=ltotal.unit, time_unit=eto_unit)

    # Compute time-of-flight of the bin edges using the interpolator
    tofs = interp(
        ltotal=ltotal.broadcast(sizes=etos.sizes),
        event_time_offset=etos,
        pulse_period=pulse_period,
    )

    return rebinned.assign_coords(tof=tofs).drop_coords(
        list({key} & {"time_of_flight", "frame_time"})
    )


def _guess_pulse_stride_offset(
    pulse_index: sc.Variable,
    ltotal: sc.Variable,
    event_time_offset: sc.Variable,
    pulse_period: sc.Variable,
    pulse_stride: int,
    interp: TofInterpolator,
) -> int:
    """
    Using the minimum ``event_time_zero`` to calculate a reference time when computing
    the time-of-flight for the neutron events makes the workflow depend on when the
    first event was recorded. There is no straightforward way to know if we started
    recording at the beginning of a frame, or half-way through a frame, without looking
    at the chopper logs. This can be manually corrected using the pulse_stride_offset
    parameter, but this makes automatic reduction of the data difficult.
    See https://github.com/scipp/essreduce/issues/184.

    Here, we perform a simple guess for the ``pulse_stride_offset`` if it is not
    provided.
    We choose a few random events, compute the time-of-flight for every possible value
    of pulse_stride_offset, and return the value that yields the least number of NaNs
    in the computed time-of-flight.

    Parameters
    ----------
    pulse_index:
        Pulse index for every event.
    ltotal:
        Total length of the flight path from the source to the detector for each event.
    event_time_offset:
        Time of arrival of the neutron at the detector for each event.
    pulse_period:
        Period of the source pulses, i.e., time between consecutive pulse starts.
    pulse_stride:
        Stride of used pulses.
    interp:
        Interpolator for the lookup table.
    """
    tofs = {}
    # Choose a few random events to compute the time-of-flight
    inds = np.random.choice(
        len(event_time_offset), min(5000, len(event_time_offset)), replace=False
    )
    pulse_index = sc.array(
        dims=pulse_index.dims,
        values=pulse_index.values[inds],
        unit=pulse_index.unit,
    )
    ltotal = sc.array(dims=ltotal.dims, values=ltotal.values[inds], unit=ltotal.unit)
    etos = sc.array(
        dims=event_time_offset.dims,
        values=event_time_offset.values[inds],
        unit=event_time_offset.unit,
    )
    for i in range(pulse_stride):
        pulse_inds = (pulse_index + i) % pulse_stride
        tofs[i] = interp(
            ltotal=ltotal,
            event_time_offset=etos,
            pulse_index=pulse_inds,
            pulse_period=pulse_period,
        )
    # Find the entry in the list with the least number of nan values
    return sorted(tofs, key=lambda x: sc.isnan(tofs[x]).sum())[0]


def _time_of_flight_data_events(
    da: sc.DataArray,
    lookup: sc.DataArray,
    ltotal: sc.Variable,
    pulse_stride_offset: int,
) -> sc.DataArray:
    etos = da.bins.coords["event_time_offset"].to(dtype=float, copy=False)
    eto_unit = elem_unit(etos)

    # Create linear interpolator
    interp = TofInterpolator(lookup, distance_unit=ltotal.unit, time_unit=eto_unit)

    # Operate on events (broadcast distances to all events)
    ltotal = sc.bins_like(etos, ltotal).bins.constituents["data"]
    etos = etos.bins.constituents["data"]

    pulse_index = None
    pulse_period = lookup.coords["pulse_period"].to(unit=eto_unit)
    pulse_stride = lookup.coords["pulse_stride"].value

    if pulse_stride > 1:
        # Compute a pulse index for every event: it is the index of the pulse within a
        # frame period. The index ranges from zero to pulse_stride - 1.
        etz_unit = 'ns'
        etz = (
            da.bins.coords["event_time_zero"]
            .bins.constituents["data"]
            .to(unit=etz_unit, copy=False)
        )
        pulse_period_ns = pulse_period.to(unit=etz_unit, dtype=int)
        frame_period = pulse_period_ns * pulse_stride
        # Define a common reference time using epoch as a base, but making sure that it
        # is aligned with the pulse_period and the frame_period.
        # We need to use a global reference time instead of simply taking the minimum
        # event_time_zero because the events may arrive in chunks, and the first event
        # may not be the first event of the first pulse for all chunks. This would lead
        # to inconsistent pulse indices.
        epoch = sc.datetime(0, unit=etz_unit)
        diff_to_epoch = (etz.min() - epoch) % pulse_period_ns
        # Here we offset the reference by half a pulse period to avoid errors from
        # fluctuations in the event_time_zeros in the data. They are triggered by the
        # neutron source, and may not always be exactly separated by the pulse period.
        # While fluctuations will exist, they will be small, and offsetting the times
        # by half a pulse period is a simple enough fix.
        reference = epoch + diff_to_epoch - (pulse_period_ns // 2)
        # Use in-place operations to avoid large allocations
        pulse_index = etz - reference
        pulse_index %= frame_period
        pulse_index //= pulse_period_ns

        # Apply the pulse_stride_offset
        if pulse_stride_offset is None:
            pulse_stride_offset = _guess_pulse_stride_offset(
                pulse_index=pulse_index,
                ltotal=ltotal,
                event_time_offset=etos,
                pulse_period=pulse_period,
                pulse_stride=pulse_stride,
                interp=interp,
            )
        pulse_index += pulse_stride_offset
        pulse_index %= pulse_stride

    # Compute time-of-flight for all neutrons using the interpolator
    tofs = interp(
        ltotal=ltotal,
        event_time_offset=etos,
        pulse_index=pulse_index,
        pulse_period=pulse_period,
    )

    parts = da.bins.constituents
    parts["data"] = tofs
    result = da.bins.assign_coords(tof=sc.bins(**parts, validate_indices=False))
    return result.bins.drop_coords("event_time_offset")


def detector_ltotal_from_straight_line_approximation(
    detector_beamline: CalibratedBeamline[RunType],
) -> DetectorLtotal[RunType]:
    """
    Compute Ltotal for the detector pixels.
    This is a naive straight-line approximation to Ltotal based on basic component
    positions.

    Parameters
    ----------
    detector_beamline:
        Beamline data for the detector that contains the positions necessary to compute
        the straight-line approximation to Ltotal (source, sample, and detector
        positions).
    """
    graph = scn.conversion.graph.beamline.beamline(scatter=True)
    return DetectorLtotal[RunType](
        detector_beamline.transform_coords(
            "Ltotal", graph=graph, keep_intermediate=False
        ).coords["Ltotal"]
    )


def monitor_ltotal_from_straight_line_approximation(
    monitor_beamline: CalibratedMonitor[RunType, MonitorType],
) -> MonitorLtotal[RunType, MonitorType]:
    """
    Compute Ltotal for the monitor.
    This is a naive straight-line approximation to Ltotal based on basic component
    positions.

    Parameters
    ----------
    monitor_beamline:
        Beamline data for the monitor that contains the positions necessary to compute
        the straight-line approximation to Ltotal (source and monitor positions).
    """
    graph = scn.conversion.graph.beamline.beamline(scatter=False)
    return MonitorLtotal[RunType, MonitorType](
        monitor_beamline.transform_coords(
            "Ltotal", graph=graph, keep_intermediate=False
        ).coords["Ltotal"]
    )


def _compute_tof_data(
    da: sc.DataArray,
    lookup: sc.DataArray,
    ltotal: sc.Variable,
    pulse_stride_offset: int,
) -> sc.DataArray:
    if da.bins is None:
        data = _time_of_flight_data_histogram(da=da, lookup=lookup, ltotal=ltotal)
        return rebin_strictly_increasing(data, dim='tof')
    else:
        return _time_of_flight_data_events(
            da=da,
            lookup=lookup,
            ltotal=ltotal,
            pulse_stride_offset=pulse_stride_offset,
        )


def detector_time_of_flight_data(
    detector_data: DetectorData[RunType],
    lookup: TimeOfFlightLookupTable,
    ltotal: DetectorLtotal[RunType],
    pulse_stride_offset: PulseStrideOffset,
) -> DetectorTofData[RunType]:
    """
    Convert the time-of-arrival data to time-of-flight data using a lookup table.
    The output data will have a time-of-flight coordinate.

    Parameters
    ----------
    da:
        Raw detector data loaded from a NeXus file, e.g., NXdetector containing
        NXevent_data.
    lookup:
        Lookup table giving time-of-flight as a function of distance and time of
        arrival.
    ltotal:
        Total length of the flight path from the source to the detector.
    pulse_stride_offset:
        When pulse-skipping, the offset of the first pulse in the stride. This is
        typically zero but can be a small integer < pulse_stride.
    """
    return DetectorTofData[RunType](
        _compute_tof_data(
            da=detector_data,
            lookup=lookup,
            ltotal=ltotal,
            pulse_stride_offset=pulse_stride_offset,
        )
    )


def monitor_time_of_flight_data(
    monitor_data: MonitorData[RunType, MonitorType],
    lookup: TimeOfFlightLookupTable,
    ltotal: MonitorLtotal[RunType, MonitorType],
    pulse_stride_offset: PulseStrideOffset,
) -> MonitorTofData[RunType, MonitorType]:
    """
    Convert the time-of-arrival data to time-of-flight data using a lookup table.
    The output data will have a time-of-flight coordinate.

    Parameters
    ----------
    da:
        Raw monitor data loaded from a NeXus file, e.g., NXmonitor containing
        NXevent_data.
    lookup:
        Lookup table giving time-of-flight as a function of distance and time of
        arrival.
    ltotal:
        Total length of the flight path from the source to the monitor.
    pulse_stride_offset:
        When pulse-skipping, the offset of the first pulse in the stride. This is
        typically zero but can be a small integer < pulse_stride.
    """
    return MonitorTofData[RunType, MonitorType](
        _compute_tof_data(
            da=monitor_data,
            lookup=lookup,
            ltotal=ltotal,
            pulse_stride_offset=pulse_stride_offset,
        )
    )


def providers() -> tuple[Callable]:
    """
    Providers of the time-of-flight workflow.
    """
    return (
        detector_time_of_flight_data,
        monitor_time_of_flight_data,
        detector_ltotal_from_straight_line_approximation,
        monitor_ltotal_from_straight_line_approximation,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""Shrink a nexus file by removing all but the first N pulses."""

import shutil
import subprocess
from pathlib import Path

import h5py as h5
import numpy as np
import scippnexus as snx

N_PULSE = 2

DIR = Path("data/coda")
IN_FNAME = "977695_00068064.hdf"
TMP_FNAME = f"TMP_{IN_FNAME}"
OUT_FNAME = f"TEST_{IN_FNAME}"


def main() -> None:
    shutil.copy(DIR / IN_FNAME, DIR / TMP_FNAME)
    with snx.File(DIR / TMP_FNAME, 'r') as f:
        entry = f['entry']
        user_names = list(entry[snx.NXuser])
        det_names = list(entry['instrument'][snx.NXdetector])
        mon_names = list(entry['instrument'][snx.NXmonitor])

    last_times = []

    with h5.File(DIR / TMP_FNAME, 'a') as f:
        entry = f['entry']
        for name in user_names:
            del entry[name]
        del entry['scripts']

        instr = entry['instrument']
        for name in det_names:
            det = instr[name]
            del det['pixel_shape']

            event_data_name = next(name for name in det if name.endswith('event_data'))
            events = det[event_data_name]

            last_times.append(
                np.array(
                    int(events['event_time_zero'][N_PULSE]),
                    dtype=f'datetime64[{events["event_time_zero"].attrs["units"]}]',
                )
            )

            n_events = events['event_index'][N_PULSE]
            slice_dataset(events, 'event_index', N_PULSE)
            slice_dataset(events, 'event_time_zero', N_PULSE)
            slice_dataset(events, 'event_id', n_events)
            slice_dataset(events, 'event_time_offset', n_events)

        last_time = np.min(last_times)
        for name in mon_names:
            mon = instr[name]
            data = mon[f'{name}_events']
            time = data['time'][()].astype(f'datetime64[{data["time"].attrs["units"]}]')
            if time.shape == (0,):
                continue
            n_times = np.argmax(time > last_time)
            slice_dataset(data, 'time', n_times)

            assert (  # noqa: S101
                data.attrs['axes'][0] == 'time'
            )  # required to slice as below:
            slice_dataset(data, 'signal', n_times)

    subprocess.check_call(['h5repack', DIR / TMP_FNAME, DIR / OUT_FNAME])  # noqa: S603, S607


def slice_dataset(group, name, n):
    attrs = group[name].attrs
    group[name] = group.pop(name)[:n]
    group[name].attrs.update(attrs)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from collections.abc import Iterable

import sciline
import scipp as sc

from ..nexus import GenericNeXusWorkflow
from . import eto_to_tof
from .types import (
    PulseStrideOffset,
    TimeOfFlightLookupTable,
    TimeOfFlightLookupTableFilename,
)


def load_tof_lookup_table(
    filename: TimeOfFlightLookupTableFilename,
) -> TimeOfFlightLookupTable:
    return TimeOfFlightLookupTable(sc.io.load_hdf5(filename))


def GenericTofWorkflow(
    *,
    run_types: Iterable[sciline.typing.Key],
    monitor_types: Iterable[sciline.typing.Key],
) -> sciline.Pipeline:
    """
    Generic workflow for computing the neutron time-of-flight for detector and monitor
    data.

    This workflow builds on the ``GenericNeXusWorkflow`` and computes time-of-flight
    from a lookup table that is created from the chopper settings, detector Ltotal and
    the neutron time-of-arrival.

    It is possible to limit which run types and monitor types
    are supported by the returned workflow.
    This is useful to reduce the size of the workflow and make it easier to inspect.
    Make sure to add *all* required run types and monitor types when using this feature.

    Attention
    ---------
    Filtering by run type and monitor type does not work with nested type vars.
    E.g., if you have a type like ``Outer[Inner[RunType]]``, this type and its
    provider will be removed.

    Parameters
    ----------
    run_types:
        List of run types to include in the workflow.
        Constrains the possible values of :class:`ess.reduce.nexus.types.RunType`.
    monitor_types:
        List of monitor types to include in the workflow.
        Constrains the possible values of :class:`ess.reduce.nexus.types.MonitorType`
        and :class:`ess.reduce.nexus.types.Component`.

    Returns
    -------
    :
        The workflow.
    """
    wf = GenericNeXusWorkflow(run_types=run_types, monitor_types=monitor_types)

    for provider in eto_to_tof.providers():
        wf.insert(provider)

    wf.insert(load_tof_lookup_table)

    # Default parameters
    wf[PulseStrideOffset] = None

    return wf



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)


import numpy as np
import scipp as sc


def find_strictly_increasing_sections(var: sc.Variable) -> list[slice]:
    """
    Find strictly increasing sections in a coordinate dimension (minimum length 2).

    Parameters
    ----------
    var:
        The variable to analyze, which should be one-dimensional.

    Returns
    -------
    sections:
        Slice objects that can be used extract strictly increasing sections.
    """
    values = var.values
    finite = np.isfinite(values)
    increasing = (np.sign(np.diff(values)) > 0) & finite[:-1] & finite[1:]
    # 1 marks the start of an increasing section, -1 marks the end
    transitions = np.diff(np.concatenate(([False], increasing, [False])).astype(int))
    section_starts = np.where(transitions == 1)[0]
    section_ends = np.where(transitions == -1)[0] + np.array(1)
    return [
        slice(start, end)
        for start, end in zip(section_starts, section_ends, strict=True)
        if end - start >= 2  # Ensure section has at least 2 points
    ]


def get_min_max(
    var: sc.Variable, *, dim: str, slices: list[slice]
) -> tuple[sc.Variable, sc.Variable]:
    if not slices:
        raise ValueError("No strictly increasing sections found.")
    combined = sc.concat([var[dim, slice] for slice in slices], dim)
    return combined.min(), combined.max()


def make_regular_grid(
    var: sc.Variable, *, dim: str, slices: list[slice]
) -> sc.Variable:
    """
    Create a regular grid variable based on the min and max of the slices.

    The grid is constructed such that it includes the minimum and maximum values
    of the strictly increasing sections, with a step size equal to the difference
    between the first two values of the section with the minimum start value (which is
    not necessarily the first section).
    """
    min_val, max_val = get_min_max(var, dim=dim, slices=slices)
    first: sc.Variable | None = None
    for s in slices:
        first = var[dim, s]
        if sc.identical(first[0], min_val):
            break
    if first is None:
        # This should not happen if slices are correctly identified and passed from
        # find_strictly_increasing_sections.
        raise ValueError("Section is not strictly increasing.")
    step = first[1] - first[0]
    return sc.arange(
        dim=dim,
        start=min_val.value,
        stop=max_val.value + step.value,  # Ensure the last bin edge is included
        step=step.value,
        unit=step.unit,
        dtype=step.dtype,
    )


def rebin_strictly_increasing(da: sc.DataArray, dim: str) -> sc.DataArray:
    """
    Find strictly monotonic sections in a coordinate dimension and rebin the data array
    into a regular grid based on these sections.
    """
    # Ensure the dimension is named like the coordinate.
    da = da.rename_dims({da.coords[dim].dim: dim})
    slices = find_strictly_increasing_sections(da.coords[dim])
    if len(slices) == 1:
        return da[dim, slices[0]]
    if not slices:
        raise ValueError("No strictly increasing sections found.")
    if da.coords[dim].dtype not in (sc.DType.float64, sc.DType.float32):
        # rebin does not like integer coords.
        da = da.assign_coords({dim: da.coords[dim].to(dtype='float64')})
    # Slices refer to the indices in the coord, which are bin edges. For slicing data
    # we need to stop at the last index minus one.
    sections = [da[dim, section.start : section.stop - 1] for section in slices]
    edges = make_regular_grid(da.coords[dim], dim=dim, slices=slices)
    return sc.reduce([sc.rebin(section, {dim: edges}) for section in sections]).sum()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

from ess.reduce.live import raw


@pytest.mark.parametrize(
    'sigma',
    [sc.scalar(0.01, unit='m'), sc.scalar(1.0, unit='cm'), sc.scalar(10.0, unit='mm')],
)
def test_gaussian_position_noise_is_sigma_unit_independent(sigma: sc.Variable) -> None:
    sigma_m = sigma.to(unit='m')
    reference = raw.gaussian_position_noise(sigma=sigma_m)
    noise = raw.gaussian_position_noise(sigma=sigma)
    assert sc.identical(noise, reference)


def test_clear_counts_resets_counts_to_zero() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.Detector(detector_number)
    assert det.data.sum().value == 0
    det.add_counts([1, 2, 3, 2])
    assert det.data.sum().value == 4
    det.clear_counts()
    assert det.data.sum().value == 0


def test_Detector_bincount_drops_out_of_range_ids() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.Detector(detector_number)
    counts = det.bincount([1, 2, 0, -1, 3, 4])
    assert sc.identical(
        counts.data,
        sc.array(dims=['pixel'], values=[1, 1, 1], unit='counts', dtype='int32'),
    )


def test_Detector_bincount_raises_if_detector_number_not_sorted() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 3, 2], unit=None)
    det = raw.Detector(detector_number)
    with pytest.raises(ValueError, match="sorted"):
        det.bincount([1])


def test_Detector_bincount_raises_if_detector_number_not_consecutive() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 4], unit=None)
    det = raw.Detector(detector_number)
    with pytest.raises(ValueError, match="consecutive"):
        det.bincount([1])


@pytest.mark.parametrize('window', [1, 2, 33])
def test_RollingDetectorView_max_window_property_returns_configured(
    window: int,
) -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=window)
    assert det.max_window == window


def test_RollingDetectorView_full_window() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=2)
    rolling = det.get()
    assert rolling.sizes == {'pixel': 3}
    assert rolling.sum().value == 0
    det.add_counts([1, 2, 3, 2])
    assert det.get().sum().value == 4
    det.add_counts([1, 3, 2])
    assert det.get().sum().value == 7
    det.add_counts([1, 2])
    assert det.get().sum().value == 5
    det.add_counts([])
    assert det.get().sum().value == 2


def test_RollingDetectorView_cumulative_returns_everything_since_clear() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=2)
    expected = sc.zeros_like(det.get())
    assert sc.identical(det.cumulative, expected)
    det.add_counts([1, 2, 3, 2])
    delta1 = det.get(window=1)
    expected += delta1
    assert sc.identical(det.cumulative, expected)
    det.add_counts([1, 3, 2])
    delta2 = det.get(window=1)
    expected += delta2
    assert sc.identical(det.cumulative, expected)
    assert sc.identical(det.cumulative, det.get())  # everything still in window
    det.add_counts([1, 2])
    delta3 = det.get(window=1)
    expected += delta3
    assert sc.identical(det.cumulative, expected)
    assert not sc.identical(det.cumulative, det.get())  # delta1 fell out of window
    det.add_counts([])
    assert sc.identical(det.cumulative, expected)
    # Reset
    det.clear_counts()
    expected = sc.zeros_like(expected)
    assert sc.identical(det.cumulative, expected)
    assert sc.identical(det.cumulative, expected)
    det.add_counts([1, 2, 3, 2])
    delta1 = det.get(window=1)
    expected += delta1
    assert sc.identical(det.cumulative, expected)


def test_RollingDetectorView_add_events_accepts_unsorted_detector_number() -> None:
    detector_number = sc.array(dims=['detector_number'], values=[1, 3, 2], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=2)
    pixel = sc.array(dims=['event'], values=[1, 2, 3, 2], unit=None)
    events = sc.DataArray(sc.ones_like(pixel), coords={'detector_number': pixel})
    det.add_events(events.group(detector_number))
    assert det.get().sum().value == 4


def test_RollingDetectorView_add_events_accepts_non_consecutive_detector_number() -> (
    None
):
    detector_number = sc.array(dims=['detector_number'], values=[1, 2, 4], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=2)
    pixel = sc.array(dims=['event'], values=[1, 2, 4, 2], unit=None)
    events = sc.DataArray(sc.ones_like(pixel), coords={'detector_number': pixel})
    det.add_events(events.group(detector_number))
    assert det.get().sum().value == 4


def test_RollingDetectorView_partial_window() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=3)
    det.add_counts([1, 2, 3, 2])
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 4
    assert det.get(window=2).sum().value == 4
    assert det.get(window=3).sum().value == 4
    det.add_counts([1, 3, 2])
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 3
    assert det.get(window=2).sum().value == 7
    assert det.get(window=3).sum().value == 7
    det.add_counts([1, 2])
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 2
    assert det.get(window=2).sum().value == 5
    assert det.get(window=3).sum().value == 9
    det.add_counts([])
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 0
    assert det.get(window=2).sum().value == 2
    assert det.get(window=3).sum().value == 5
    det.add_counts([1, 2])
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 2
    assert det.get(window=2).sum().value == 2
    assert det.get(window=3).sum().value == 4


def test_RollingDetectorView_clear_counts() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=3)
    det.add_counts([1, 2, 3, 2])
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 4
    assert det.get(window=2).sum().value == 4
    assert det.get(window=3).sum().value == 4
    assert det.get().sum().value == 4
    det.clear_counts()
    assert det.get(window=0).sum().value == 0
    assert det.get(window=1).sum().value == 0
    assert det.get(window=2).sum().value == 0
    assert det.get(window=3).sum().value == 0
    assert det.get().sum().value == 0


def test_RollingDetectorView_raises_if_subwindow_exceeds_window() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=3)
    with pytest.raises(ValueError, match="Window size"):
        det.get(window=4)
    with pytest.raises(ValueError, match="Window size"):
        det.get(window=-1)


def test_RollingDetectorView_projects_counts() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    # Dummy projection that just drops the first pixel
    det = raw.RollingDetectorView(
        detector_number=detector_number,
        window=3,
        projection=lambda da: da['pixel', 1:].rename(pixel='abc'),
    )
    expected = sc.DataArray(
        sc.array(dims=['pixel'], values=[0, 0], unit='counts', dtype='int32'),
        coords={'detector_number': detector_number[1:]},
    ).rename(pixel='abc')

    det.add_counts([1, 2, 3, 2])
    expected.values = [2, 1]
    assert sc.identical(det.get(), expected)

    det.add_counts([1, 3, 3, 1])
    expected.values = [2, 3]
    assert sc.identical(det.get(), expected)


def test_project_xy_with_given_zplane_scales() -> None:
    result = raw.project_xy(
        sc.vectors(
            dims=['point'],
            values=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [6.0, 10.0, 12.0]],
            unit='m',
        ),
        zplane=sc.scalar(6.0, unit='m'),
    )
    assert sc.identical(
        result,
        sc.DataGroup(
            x=sc.array(dims=['point'], values=[2.0, 4.0, 3.0], unit='m'),
            y=sc.array(dims=['point'], values=[4.0, 5.0, 5.0], unit='m'),
            z=sc.scalar(6.0, unit='m'),
        ),
    )


def test_project_xy_defaults_to_scale_to_zmin() -> None:
    result = raw.project_xy(
        sc.vectors(
            dims=['point'],
            values=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [6.0, 10.0, 12.0]],
            unit='m',
        )
    )
    assert sc.identical(
        result,
        sc.DataGroup(
            # Note same relative values as with zplane=6, just scaled.
            x=sc.array(dims=['point'], values=[1.0, 2.0, 1.5], unit='m'),
            y=sc.array(dims=['point'], values=[2.0, 2.5, 2.5], unit='m'),
            z=sc.scalar(3.0, unit='m'),
        ),
    )


def test_project_onto_cylinder_z() -> None:
    radius = sc.scalar(2.0, unit='m')
    # Input radii are 4 and 1 => scale by 1/2 and 2.
    result = raw.project_onto_cylinder_z(
        sc.vectors(dims=['point'], values=[[0.0, 4.0, 3.0], [1.0, 0.0, 6.0]], unit='m'),
        radius=radius,
    )
    assert sc.identical(result['r'], radius)
    assert sc.identical(
        result['z'], sc.array(dims=['point'], values=[1.5, 12.0], unit='m')
    )
    assert sc.identical(
        result['phi'], sc.array(dims=['point'], values=[90.0, 0.0], unit='deg')
    )
    assert sc.identical(
        result['arc_length'],
        sc.array(dims=['point'], values=[radius.value * np.pi * 0.5, 0.0], unit='m'),
    )


def make_grid_cube(
    nx: int = 5,
    ny: int = 5,
    nz: int = 5,
    center: tuple = (0.0, 0.0, 10.0),
    size: float = 1.0,
) -> sc.Variable:
    """Create a grid of points in a cube centered at specified position.

    Parameters
    ----------
    nx:
        Number of points along x-axis, by default 5
    ny:
        Number of points along y-axis, by default 5
    nz:
        Number of points along z-axis, by default 5
    center:
        (x, y, z) coordinates of cube center, by default (0.0, 0.0, 10.0)
    size:
        Side length of cube in meters, by default 1.0

    Returns
    -------
    :
        Scipp variable containing grid points with shape (nx * ny * nz, 3)

    Examples
    --------
    >>> grid = make_grid_cube(nx=3, ny=3, nz=3)
    >>> grid.shape
    (27, 3)
    """
    # Create coordinate arrays
    x = np.linspace(-size / 2, size / 2, nx) + center[0]
    y = np.linspace(-size / 2, size / 2, ny) + center[1]
    z = np.linspace(-size / 2, size / 2, nz) + center[2]

    # Create meshgrid
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    # Stack into points array
    points = np.stack([X.flatten(), Y.flatten(), Z.flatten()]).T

    return sc.vectors(dims=['point'], values=points, unit='m')


def test_histogrammer_input_indices() -> None:
    nx, ny, nz = 3, 3, 3
    coords = raw.project_xy(
        make_grid_cube(nx=nx, ny=ny, nz=nz, center=(0.0, 3.0, 10.0))
    )
    coords = sc.concat([coords], 'replica')

    resolution = {'x': 4, 'y': 5}
    histogrammer = raw.Histogrammer.from_coords(coords=coords, resolution=resolution)
    indices = histogrammer.input_indices()
    assert set(indices.coords) == {'x', 'y'}
    assert indices.bins.size().sum().value == nx * ny * nz
    assert indices.sizes == resolution


def test_ROIFilter_from_trivial_RollingDetectorView() -> None:
    detector_number = sc.array(
        dims=['x', 'y'], values=[[1, 2, 3], [4, 5, 6]], unit=None
    )
    view = raw.RollingDetectorView(detector_number=detector_number, window=2)
    roi_filter = view.make_roi_filter()
    data = detector_number.copy()
    data.unit = 'counts'
    flat = data.flatten(to='detector_number')

    result, scale = roi_filter.apply(data)
    # ROIFilter defaults to include nothing
    assert sc.identical(result, flat[0:0])
    assert sc.identical(scale, sc.zeros(dims=['detector_number'], shape=[0]))

    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(1, 2)))
    result, scale = roi_filter.apply(data)
    assert sc.identical(result, flat[3:6])
    assert sc.identical(scale, sc.ones(dims=['detector_number'], shape=[3]))

    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(1, 2), y=(1, 3)))
    result, scale = roi_filter.apply(data)
    assert sc.identical(result, flat[4:6])
    assert sc.identical(scale, sc.ones(dims=['detector_number'], shape=[2]))


def test_ROIFilter_from_RollingDetectorView_with_custom_projection() -> None:
    detector_number = sc.array(
        dims=['x', 'y', 'z'], values=[[[1, 2], [3, 4]], [[5, 6], [7, 8]]], unit=None
    )
    view = raw.RollingDetectorView(
        detector_number=detector_number, window=2, projection=lambda da: da['z', 0]
    )
    roi_filter = view.make_roi_filter()
    data = detector_number.copy()
    data.unit = 'counts'
    flat = data['z', 0].flatten(to='detector_number')

    result, scale = roi_filter.apply(data)
    # ROIFilter defaults to include nothing
    assert sc.identical(result, flat[0:0])
    assert sc.identical(scale, sc.zeros(dims=['detector_number'], shape=[0]))

    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(1, 2)))
    result, scale = roi_filter.apply(data)
    assert sc.identical(result, flat[2:4])
    assert sc.identical(scale, sc.ones(dims=['detector_number'], shape=[2]))

    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(1, 2), y=(1, 3)))
    result, scale = roi_filter.apply(data)
    assert sc.identical(result, flat[3:4])
    assert sc.identical(scale, sc.ones(dims=['detector_number'], shape=[1]))


def test_ROIFilter_from_RollingDetectorView_with_xy_projection() -> None:
    detector_number = sc.array(
        dims=['x', 'y', 'z'], values=[[[1, 2], [3, 4]], [[5, 6], [7, 8]]], unit=None
    )
    nx, ny, nz = 2, 2, 2
    coords = raw.project_xy(
        make_grid_cube(nx=nx, ny=ny, nz=nz, center=(0.0, 0.0, 10.0))
    )
    coords = sc.concat(
        [coords.fold(dim='point', sizes=detector_number.sizes)], 'replica'
    )

    resolution = {'x': 4, 'y': 4}
    histogrammer = raw.Histogrammer.from_coords(coords=coords, resolution=resolution)
    view = raw.RollingDetectorView(
        detector_number=detector_number, window=1, projection=histogrammer
    )
    roi_filter = view.make_roi_filter()
    data = detector_number.copy()
    data.unit = 'counts'
    flat = data.flatten(to='detector_number')

    result, scale = roi_filter.apply(data)
    # ROIFilter defaults to include nothing
    assert sc.identical(result, flat[0:0])
    assert sc.identical(scale, sc.zeros(dims=['detector_number'], shape=[0]))

    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(0, 2)))
    result, scale = roi_filter.apply(data)
    assert sc.identical(scale, sc.ones(dims=['detector_number'], shape=[4]))
    assert sc.identical(result, flat[:4])

    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(0, 2), y=(0, 2)))
    result, scale = roi_filter.apply(data)
    assert sc.identical(scale, sc.ones(dims=['detector_number'], shape=[2]))
    assert sc.identical(result, flat[0:2])


def test_Histogrammer_given_empty_bins_returns_nans_in_transform_weights() -> None:
    x = sc.array(dims=['replica', 'pixel'], values=[[0.0, 1.0]], unit='m')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    resolution = {'y': 1, 'x': 4}
    view = raw.RollingDetectorView(
        detector_number=sc.array(dims=['pixel'], values=[1, 2], unit=None),
        window=1,
        projection=raw.Histogrammer.from_coords(coords=coords, resolution=resolution),
    )
    weights = view.transform_weights()
    assert_identical(
        weights.data,
        sc.array(dims=['y', 'x'], values=[[1.0, np.nan, np.nan, 1.0]], dtype='float32'),
    )


def test_Histogrammer_threshold_based_on_non_empty_bins_in_transform_weights() -> None:
    x = sc.array(dims=['replica', 'pixel'], values=[[0.0, 0.0, 1.0]], unit='m')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    resolution = {'y': 1, 'x': 4}
    view = raw.RollingDetectorView(
        detector_number=sc.array(dims=['pixel'], values=[1, 2, 3], unit=None),
        window=1,
        projection=raw.Histogrammer.from_coords(coords=coords, resolution=resolution),
    )
    weights = view.transform_weights(threshold=0.9)
    assert_identical(
        weights.data,
        sc.array(
            dims=['y', 'x'], values=[[2.0, np.nan, np.nan, np.nan]], dtype='float32'
        ),
    )


def test_Histogrammer_replicas_do_not_affect_transformed_weights() -> None:
    resolution = {'y': 1, 'x': 4}
    x = sc.array(dims=['replica', 'pixel'], values=[[0.0, 0.0, 0.5, 1.0]], unit='m')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3, 4], unit=None)
    view = raw.RollingDetectorView(
        detector_number=detector_number,
        window=1,
        projection=raw.Histogrammer.from_coords(coords=coords, resolution=resolution),
    )
    reference = view.transform_weights()

    x = sc.concat([x, x], 'replica')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    view = raw.RollingDetectorView(
        detector_number=detector_number,
        window=1,
        projection=raw.Histogrammer.from_coords(coords=coords, resolution=resolution),
    )
    result = view.transform_weights()

    assert_identical(reference, result)


def test_RollingDetectorView_normalization() -> None:
    resolution = {'y': 1, 'x': 4}
    x = sc.array(dims=['replica', 'pixel'], values=[[0.0, 0.0, 0.3, 1.0]], unit='m')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    histogrammer = raw.Histogrammer.from_coords(coords=coords, resolution=resolution)
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3, 4], unit=None)
    view = raw.RollingDetectorView(
        detector_number=detector_number, window=1, projection=histogrammer
    )
    weights = view.transform_weights().data
    expected = sc.array(
        dims=['y', 'x'],
        values=[[0.0, 0.0, np.nan, 0.0]],
        unit='counts',
        dtype='float32',
    )
    assert_identical(view.get().data / weights, expected)
    assert_identical(view.get().data / weights, expected)
    expected.values = [[0.0, 0.0, 0.0, 0.0]]
    assert_identical(view.get().data, expected.to(dtype='int32'))

    # detector_number 1 and 2 are in the same bin, so their counts are summed,
    view.add_counts([1, 2, 3, 4])
    expected.values = [[1.0, 1.0, np.nan, 1.0]]
    assert_identical(view.get().data / weights, expected)
    assert_identical(view.get().data / weights, expected)
    expected.values = [[2.0, 1.0, 0.0, 1.0]]
    assert_identical(view.get().data, expected.to(dtype='int32'))

    view.add_counts([2, 4])
    expected.values = [[0.5, 0.0, np.nan, 1.0]]
    assert_identical(view.get().data / weights, expected)
    assert_identical(view.get().data / weights, expected)
    expected.values = [[1.0, 0.0, 0.0, 1.0]]
    assert_identical(view.get().data, expected.to(dtype='int32'))


def test_transform_weights_raises_given_bad_sizes() -> None:
    resolution = {'y': 1, 'x': 4}
    x = sc.array(dims=['replica', 'pixel'], values=[[0.0, 0.0, 0.3, 1.0]], unit='m')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    histogrammer = raw.Histogrammer.from_coords(coords=coords, resolution=resolution)
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3, 4], unit=None)
    view = raw.RollingDetectorView(
        detector_number=detector_number, window=1, projection=histogrammer
    )
    with pytest.raises(sc.DimensionError):
        view.transform_weights(sc.ones(dims=['x'], shape=[4]))
    with pytest.raises(sc.DimensionError):
        view.transform_weights(sc.ones(dims=['pixel'], shape=[3]))


def test_transform_weights_raises_given_DataArray_with_bad_det_num() -> None:
    resolution = {'y': 1, 'x': 4}
    x = sc.array(dims=['replica', 'pixel'], values=[[0.0, 0.0, 1.0]], unit='m')
    coords = sc.DataGroup(x=x, y=x, z=sc.zeros_like(x))
    histogrammer = raw.Histogrammer.from_coords(coords=coords, resolution=resolution)
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    view = raw.RollingDetectorView(
        detector_number=detector_number, window=1, projection=histogrammer
    )
    weights = sc.DataArray(
        data=sc.ones(dims=['pixel'], shape=[3]),
        coords={
            'detector_number': sc.array(dims=['pixel'], values=[1, 2, 4], unit=None)
        },
    )
    with pytest.raises(sc.CoordError):
        view.transform_weights(weights)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from typing import NewType

import pytest
import sciline
import scipp as sc

from ess.reduce import streaming

DynamicA = NewType('DynamicA', float)
DynamicB = NewType('DynamicB', float)
DynamicC = NewType('DynamicC', float)
StaticA = NewType('StaticA', float)
AccumA = NewType('AccumA', float)
AccumB = NewType('AccumB', float)
AccumC = NewType('AccumC', float)
Target = NewType('Target', float)


def make_static_a() -> StaticA:
    make_static_a.call_count += 1
    return StaticA(2.0)


make_static_a.call_count = 0


def make_accum_a(value: DynamicA, static: StaticA) -> AccumA:
    return AccumA(value * static)


def make_accum_b(value: DynamicB) -> AccumB:
    return AccumB(value)


def make_accum_c(value: DynamicC) -> AccumC:
    return AccumC(value)


def make_target(accum_a: AccumA, accum_b: AccumB) -> Target:
    return Target(accum_a / accum_b)


def test_StreamProcessor_with_empty_workflow_and_targets() -> None:
    workflow = sciline.Pipeline()
    streaming_wf = streaming.StreamProcessor(
        base_workflow=workflow,
        dynamic_keys=(int,),
        target_keys=(),
        accumulators=(),
    )
    streaming_wf.accumulate({int: 1})
    result = streaming_wf.finalize()
    assert result == {}


def test_StreamProcessor_overall_behavior() -> None:
    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )
    orig_workflow = base_workflow.copy()

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 9.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 15.0))
    assert make_static_a.call_count == 1

    # Consistency check: Run the original workflow with the same inputs, all at once
    orig_workflow[DynamicA] = sc.scalar(1 + 2 + 3)
    orig_workflow[DynamicB] = sc.scalar(4 + 5 + 6)
    expected = orig_workflow.compute(Target)
    assert sc.identical(expected, result[Target])


def test_StreamProcessor_uses_custom_accumulator() -> None:
    class Always42Accumulator(streaming.Accumulator[sc.Variable]):
        def _do_push(self, value: sc.Variable) -> None:
            pass

        @property
        def is_empty(self) -> bool:
            return False

        def _get_value(self) -> sc.Variable:
            return sc.scalar(42)

        def clear(self) -> None:
            # Nothing to clear
            pass

    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators={
            AccumA: streaming.EternalAccumulator(),
            AccumB: Always42Accumulator(),
        },
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 42.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 42.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 42.0))


def test_StreamProcessor_does_not_compute_unused_static_nodes() -> None:
    def a_independent_target(accum_b: AccumB) -> Target:
        return Target(1.5 * accum_b)

    def derived_static_a(x: float) -> StaticA:
        derived_static_a.call_count += 1
        return StaticA(2.0 * x)

    derived_static_a.call_count = 0

    base_workflow = sciline.Pipeline(
        (derived_static_a, make_accum_a, make_accum_b, a_independent_target)
    )

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(1.5 * 4.0))
    assert derived_static_a.call_count == 0


def test_StreamProcess_with_zero_accumulators_for_buffered_workflow_calls() -> None:
    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )
    make_static_a.call_count = 0

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(),
        allow_bypass=True,
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 2.0 / 5.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 6.0))
    assert make_static_a.call_count == 1


def test_StreamProcessor_with_bypass() -> None:
    def _make_static_a() -> StaticA:
        _make_static_a.call_count += 1
        return StaticA(2.0)

    _make_static_a.call_count = 0

    base_workflow = sciline.Pipeline(
        (_make_static_a, make_accum_a, make_accum_b, make_target)
    )
    orig_workflow = base_workflow.copy()

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA,),  # Note: No AccumB
        allow_bypass=True,
    )
    streaming_wf.accumulate({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    streaming_wf.accumulate({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    result = streaming_wf.finalize()
    # Note denominator is 5, not 9
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 5.0))
    streaming_wf.accumulate({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    result = streaming_wf.finalize()
    # Note denominator is 6, not 15
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 6.0))
    assert _make_static_a.call_count == 1

    # Consistency check: Run the original workflow with the same inputs, all at once
    orig_workflow[DynamicA] = sc.scalar(1 + 2 + 3)
    orig_workflow[DynamicB] = sc.scalar(6)
    expected = orig_workflow.compute(Target)
    assert sc.identical(expected, result[Target])


def test_StreamProcessor_without_bypass_raises() -> None:
    def _make_static_a() -> StaticA:
        _make_static_a.call_count += 1
        return StaticA(2.0)

    _make_static_a.call_count = 0

    base_workflow = sciline.Pipeline(
        (_make_static_a, make_accum_a, make_accum_b, make_target)
    )

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA,),  # Note: No AccumB
    )
    streaming_wf.accumulate({DynamicA: 1, DynamicB: 4})
    # Sciline passes `None` to the provider that needs AccumB.
    with pytest.raises(TypeError, match='unsupported operand type'):
        _ = streaming_wf.finalize()


def test_StreamProcessor_calls_providers_after_accumulators_only_when_finalizing() -> (
    None
):
    def _make_target(accum_a: AccumA, accum_b: AccumB) -> Target:
        _make_target.call_count += 1
        return Target(accum_a / accum_b)

    _make_target.call_count = 0

    base_workflow = sciline.Pipeline(
        (make_accum_a, make_accum_b, _make_target), params={StaticA: 2.0}
    )

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    streaming_wf.accumulate({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    streaming_wf.accumulate({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert _make_target.call_count == 0
    result = streaming_wf.finalize()
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 9.0))
    assert _make_target.call_count == 1
    streaming_wf.accumulate({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert _make_target.call_count == 1
    result = streaming_wf.finalize()
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 15.0))
    assert _make_target.call_count == 2
    result = streaming_wf.finalize()
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 15.0))
    # Outputs are not cached.
    assert _make_target.call_count == 3


def test_StreamProcessor_does_not_reuse_dynamic_keys() -> None:
    base_workflow = sciline.Pipeline(
        (make_accum_a, make_accum_b, make_target), params={StaticA: 2.0}
    )
    orig_workflow = base_workflow.copy()

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2)})  # Only A
    assert not sc.identical(result[Target], sc.scalar(2 * 3.0 / 8.0))
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicB: sc.scalar(5)})  # Only B
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 9.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 15.0))

    # Consistency check: Run the original workflow with the same inputs, all at once
    orig_workflow[DynamicA] = sc.scalar(1 + 2 + 3)
    orig_workflow[DynamicB] = sc.scalar(4 + 5 + 6)
    expected = orig_workflow.compute(Target)
    assert sc.identical(expected, result[Target])


def test_StreamProcessor_raises_given_partial_update_for_accumulator() -> None:
    base_workflow = sciline.Pipeline(
        (make_accum_a, make_accum_b, make_accum_c, make_target), params={StaticA: 2.0}
    )
    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB, DynamicC),
        target_keys=(Target, AccumC),
        accumulators=(Target, AccumC),  # Target depends on both A and B
    )
    # We can update either (A, B) and/or C...
    streaming_wf.accumulate({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    streaming_wf.accumulate({DynamicC: sc.scalar(11)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    assert sc.identical(result[AccumC], sc.scalar(11))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * (1.0 / 4.0 + 2.0 / 5.0)))
    assert sc.identical(result[AccumC], sc.scalar(11))
    result = streaming_wf.add_chunk({DynamicC: sc.scalar(12)})
    assert sc.identical(result[Target], sc.scalar(2 * (1.0 / 4.0 + 2.0 / 5.0)))
    assert sc.identical(result[AccumC], sc.scalar(23))
    # ... but not just A or B
    with pytest.raises(
        ValueError,
        match=r'{tests.streaming_test.DynamicB} not provided in the current chunk',
    ):
        result = streaming_wf.add_chunk({DynamicA: sc.scalar(2)})  # Only A
    with pytest.raises(
        ValueError,
        match=r'{tests.streaming_test.DynamicA} not provided in the current chunk',
    ):
        result = streaming_wf.add_chunk({DynamicB: sc.scalar(5)})  # Only B


def test_StreamProcessor_raises_when_trying_to_update_non_dynamic_key() -> None:
    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )
    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )

    # Regular update ok
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))

    # Non-dynamic input key
    with pytest.raises(
        ValueError,
        match=r'Got non-dynamic keys: {tests.streaming_test.StaticA}',
    ):
        result = streaming_wf.add_chunk({StaticA: sc.scalar(2)})
    # Intermediate key depending on dynamic key
    with pytest.raises(
        ValueError,
        match=r'Got non-dynamic keys: {tests.streaming_test.AccumA}',
    ):
        result = streaming_wf.add_chunk({AccumA: sc.scalar(2)})
    # Target key depending on dynamic key
    with pytest.raises(
        ValueError,
        match=r'Got non-dynamic keys: {tests.streaming_test.Target}',
    ):
        result = streaming_wf.add_chunk({Target: sc.scalar(2)})


def test_StreamProcessor_clear() -> None:
    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )

    # Reset call counter to ensure we can track it properly
    make_static_a.call_count = 0

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    # Add some data
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 9.0))

    # Make sure static_a was called exactly once
    assert make_static_a.call_count == 1

    # Clear and verify we get back to initial state
    streaming_wf.clear()
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))

    # Static values should be preserved after clear, so call_count remains 1
    assert make_static_a.call_count == 1


def test_StreamProcessor_with_context() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    Static = NewType('Static', int)
    ProcessedContext = NewType('ProcessedContext', int)
    ProcessedStreamed = NewType('ProcessedStreamed', int)
    Output = NewType('Output', int)

    def make_static() -> Static:
        make_static.call_count += 1
        return Static(2)

    make_static.call_count = 0

    def process_context(context: Context, static: Static) -> ProcessedContext:
        process_context.call_count += 1
        return ProcessedContext(context * static)

    process_context.call_count = 0

    def process_streamed(
        streamed: Streamed, context: ProcessedContext
    ) -> ProcessedStreamed:
        return ProcessedStreamed(streamed + context)

    def finalize(streamed: ProcessedStreamed) -> Output:
        return Output(streamed)

    wf = sciline.Pipeline((make_static, process_context, process_streamed, finalize))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(Context,),
        target_keys=(Output,),
        accumulators=(ProcessedStreamed,),
    )
    assert make_static.call_count == 1
    assert process_context.call_count == 0
    streaming_wf.set_context({Context: sc.scalar(3)})
    assert process_context.call_count == 1
    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    assert sc.identical(streaming_wf.finalize()[Output], sc.scalar(2 * 3 + 4))
    assert make_static.call_count == 1
    assert process_context.call_count == 1
    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    assert process_context.call_count == 1
    streaming_wf.set_context({Context: sc.scalar(1)})
    # Context changed but no chunk added, so new context is not visible in the result
    assert sc.identical(streaming_wf.finalize()[Output], sc.scalar(2 * (2 * 3 + 4)))
    # The third chunk sees the context value of "1" (instead of "3")
    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    assert sc.identical(
        streaming_wf.finalize()[Output], sc.scalar(2 * (2 * 3 + 4) + 2 * 1 + 4)
    )
    assert make_static.call_count == 1
    assert process_context.call_count == 2


def test_StreamProcessor_with_context_updates_non_dynamic_targets() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    Static = NewType('Static', int)
    ProcessedContext = NewType('ProcessedContext', int)
    ProcessedStreamed = NewType('ProcessedStreamed', int)
    Output = NewType('Output', int)

    def make_static() -> Static:
        make_static.call_count += 1
        return Static(2)

    make_static.call_count = 0

    def process_context(context: Context, static: Static) -> ProcessedContext:
        process_context.call_count += 1
        return ProcessedContext(context * static)

    process_context.call_count = 0

    def process_streamed(streamed: Streamed) -> ProcessedStreamed:
        return ProcessedStreamed(streamed)

    def finalize(streamed: ProcessedStreamed) -> Output:
        return Output(streamed)

    wf = sciline.Pipeline((make_static, process_context, process_streamed, finalize))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(Context,),
        target_keys=(ProcessedContext, Output),
        accumulators=(ProcessedStreamed,),
    )
    assert make_static.call_count == 1
    assert process_context.call_count == 0

    streaming_wf.set_context({Context: sc.scalar(3)})
    assert process_context.call_count == 1
    streaming_wf.accumulate({Streamed: sc.scalar(4)})

    results = streaming_wf.finalize()
    assert sc.identical(results[Output], sc.scalar(4))
    assert sc.identical(results[ProcessedContext], sc.scalar(2 * 3))
    assert make_static.call_count == 1
    assert process_context.call_count == 1

    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    assert process_context.call_count == 1

    results = streaming_wf.finalize()
    assert process_context.call_count == 1

    streaming_wf.set_context({Context: sc.scalar(5)})
    results = streaming_wf.finalize()
    assert sc.identical(results[Output], sc.scalar(8))
    assert sc.identical(results[ProcessedContext], sc.scalar(2 * 5))
    assert process_context.call_count == 2


def test_StreamProcessor_raises_when_invalid_context_key_provided() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    NonContext = NewType('NonContext', int)
    Output = NewType('Output', int)

    def process(streamed: Streamed, context: Context) -> Output:
        return Output(streamed + context)

    wf = sciline.Pipeline((process,))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(Context,),
        target_keys=(Output,),
        accumulators=(Output,),
    )

    streaming_wf.set_context({Context: sc.scalar(3)})
    # NonContext is not in context_keys
    with pytest.raises(ValueError, match=r"Key .* is not a context key"):
        streaming_wf.set_context({NonContext: sc.scalar(5)})


def test_StreamProcessor_context_only_recomputes_descendants() -> None:
    Streamed = NewType('Streamed', int)
    ContextInputA = NewType('ContextInputA', int)
    ContextInputB = NewType('ContextInputB', int)
    ProcessedA = NewType('ProcessedA', int)
    ProcessedB = NewType('ProcessedB', int)
    Output = NewType('Output', int)

    def process_a(context: ContextInputA) -> ProcessedA:
        process_a.call_count += 1
        return ProcessedA(context * 2)

    process_a.call_count = 0

    def process_b(context: ContextInputB) -> ProcessedB:
        process_b.call_count += 1
        return ProcessedB(context * 3)

    process_b.call_count = 0

    def combine(streamed: Streamed, a: ProcessedA, b: ProcessedB) -> Output:
        return Output(streamed + a + b)

    wf = sciline.Pipeline((process_a, process_b, combine))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(ContextInputA, ContextInputB),
        target_keys=(Output,),
        accumulators=(Output,),
    )

    assert process_a.call_count == 0
    assert process_b.call_count == 0

    # Set both contexts initially
    streaming_wf.set_context({ContextInputA: sc.scalar(1), ContextInputB: sc.scalar(2)})
    assert process_a.call_count == 1
    assert process_b.call_count == 1

    # Update only ContextInputA
    streaming_wf.set_context({ContextInputA: sc.scalar(3)})
    assert process_a.call_count == 2
    assert process_b.call_count == 1  # Not called again

    # Update only ContextInputB
    streaming_wf.set_context({ContextInputB: sc.scalar(4)})
    assert process_a.call_count == 2  # Not called again
    assert process_b.call_count == 2


def test_StreamProcessor_context_retained_after_clear() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    ProcessedContext = NewType('ProcessedContext', int)
    Output = NewType('Output', int)

    def process_context(context: Context) -> ProcessedContext:
        process_context.call_count += 1
        return ProcessedContext(context * 2)

    process_context.call_count = 0

    def combine(streamed: Streamed, context: ProcessedContext) -> Output:
        return Output(streamed + context)

    wf = sciline.Pipeline((process_context, combine))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(Context,),
        target_keys=(Output,),
        accumulators=(Output,),
    )

    streaming_wf.set_context({Context: sc.scalar(5)})
    assert process_context.call_count == 1

    streaming_wf.accumulate({Streamed: sc.scalar(3)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(3 + 5 * 2))

    streaming_wf.clear()
    # Context should be preserved after clear
    streaming_wf.accumulate({Streamed: sc.scalar(3)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(3 + 5 * 2))
    # process_context shouldn't have been called again
    assert process_context.call_count == 1


def test_StreamProcessor_empty_context_update() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    Output = NewType('Output', int)

    def process(streamed: Streamed, context: Context) -> Output:
        return Output(streamed + context)

    wf = sciline.Pipeline((process,))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(Context,),
        target_keys=(Output,),
        accumulators=(Output,),
    )

    # Empty context update should not raise
    streaming_wf.set_context({})  # No-op

    streaming_wf.set_context({Context: sc.scalar(3)})
    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(7))


def test_StreamProcessor_with_overlapping_context_and_dynamic_keys_raises() -> None:
    Key = NewType('Key', int)  # Used as both context and dynamic
    Output = NewType('Output', int)

    def double(value: Key) -> Output:
        return Output(value * 2)

    wf = sciline.Pipeline((double,))

    # Using a key as both context and dynamic should raise
    with pytest.raises(ValueError, match="Keys cannot be both dynamic and context"):
        streaming.StreamProcessor(
            base_workflow=wf,
            dynamic_keys=(Key,),
            context_keys=(Key,),
            target_keys=(Output,),
            accumulators=(Output,),
        )


def test_StreamProcessor_with_unused_context_key() -> None:
    Streamed = NewType('Streamed', int)
    UsedContext = NewType('UsedContext', int)
    UnusedContext = NewType('UnusedContext', int)
    Output = NewType('Output', int)

    def process(streamed: Streamed, context: UsedContext) -> Output:
        return Output(streamed + context)

    wf = sciline.Pipeline((process,))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(UsedContext, UnusedContext),  # UnusedContext is never used
        target_keys=(Output,),
        accumulators=(Output,),
    )

    # Setting only the used context should work fine
    streaming_wf.set_context({UsedContext: sc.scalar(3)})
    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(7))

    # Setting the unused context should also work without errors
    streaming_wf.set_context({UnusedContext: sc.scalar(99)})
    streaming_wf.accumulate({Streamed: sc.scalar(5)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(7 + 8))


def test_StreamProcessor_with_separate_context_nodes() -> None:
    """Test that non-interdependent context keys work fine."""
    Streamed = NewType('Streamed', int)
    ContextInputA = NewType('ContextInputA', int)
    ContextInputB = NewType('ContextInputB', int)
    Output = NewType('Output', int)

    def process(streamed: Streamed, a: ContextInputA, b: ContextInputB) -> Output:
        return Output(streamed + a + b)

    wf = sciline.Pipeline((process,))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(ContextInputA, ContextInputB),  # Independent contexts
        target_keys=(Output,),
        accumulators=(Output,),
    )

    # Setting both contexts should work fine
    streaming_wf.set_context({ContextInputA: sc.scalar(3), ContextInputB: sc.scalar(5)})
    streaming_wf.accumulate({Streamed: sc.scalar(4)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(4 + 3 + 5))

    # Updating only one context should also work
    streaming_wf.set_context({ContextInputA: sc.scalar(7)})
    streaming_wf.accumulate({Streamed: sc.scalar(2)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(4 + 3 + 5 + 2 + 7 + 5))


def test_StreamProcessor_with_non_scalar_context() -> None:
    Streamed = NewType('Streamed', float)
    VectorContext = NewType('VectorContext', sc.Variable)
    Output = NewType('Output', float)

    def process(streamed: Streamed, context: VectorContext) -> Output:
        # Use the sum of the context vector
        return Output(streamed + sc.sum(context).value)

    wf = sciline.Pipeline((process,))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(VectorContext,),
        target_keys=(Output,),
        accumulators=(Output,),
    )

    # Set a vector context
    vector_data = sc.array(dims=['x'], values=[1.0, 2.0, 3.0])
    streaming_wf.set_context({VectorContext: vector_data})
    streaming_wf.accumulate({Streamed: sc.scalar(4.0)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(4.0 + 6.0))  # 6.0 is sum of vector

    # Update vector context
    new_vector = sc.array(dims=['x'], values=[2.0, 3.0, 4.0])
    streaming_wf.set_context({VectorContext: new_vector})
    streaming_wf.accumulate({Streamed: sc.scalar(5.0)})
    result = streaming_wf.finalize()
    assert sc.identical(result[Output], sc.scalar(4.0 + 6.0 + 5.0 + 9.0))


def test_StreamProcessor_with_invalid_context_value() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    Output = NewType('Output', int)

    def process(streamed: Streamed, context: Context) -> Output:
        # This will fail if context is not an integer
        return Output(streamed + context)

    wf = sciline.Pipeline((process,))
    streaming_wf = streaming.StreamProcessor(
        base_workflow=wf,
        dynamic_keys=(Streamed,),
        context_keys=(Context,),
        target_keys=(Output,),
        accumulators=(Output,),
    )

    # Setting an invalid context type should trigger error when using it.
    streaming_wf.set_context({Context: "not_an_integer"})
    with pytest.raises(TypeError):
        streaming_wf.accumulate({Streamed: sc.scalar(4)})


def test_StreamProcessor_rejects_interdependent_dynamic_keys() -> None:
    Streamed1 = NewType('Streamed1', int)
    Streamed2 = NewType('Streamed2', int)
    Output = NewType('Output', int)

    def process_derived_dynamic(primary: Streamed1) -> Streamed2:
        return Streamed2(primary * 10)

    def process(value1: Streamed1, value2: Streamed2) -> Output:
        return Output(value1 + value2)

    wf = sciline.Pipeline((process_derived_dynamic, process))

    # Using interdependent dynamic keys should raise a clear error
    with pytest.raises(
        ValueError,
        match=r"Dynamic key .* depends on other dynamic/context keys: .*. "
        "This is not supported.",
    ):
        streaming.StreamProcessor(
            base_workflow=wf,
            dynamic_keys=(Streamed1, Streamed2),  # Streamed2 depends on Streamed1
            target_keys=(Output,),
            accumulators=(Output,),
        )


def test_StreamProcessor_rejects_interdependent_context_keys() -> None:
    Streamed = NewType('Streamed', int)
    PrimaryContext = NewType('PrimaryContext', int)
    DerivedContext = NewType('DerivedContext', int)
    Output = NewType('Output', int)

    def process_derived_context(primary: PrimaryContext) -> DerivedContext:
        return DerivedContext(primary * 10)

    def process(streamed: Streamed, derived: DerivedContext) -> Output:
        return Output(streamed + derived)

    wf = sciline.Pipeline((process_derived_context, process))

    # Context key depending on another context key should raise
    with pytest.raises(
        ValueError,
        match=r"Context key .* depends on other dynamic/context keys: .*. "
        "This is not supported.",
    ):
        streaming.StreamProcessor(
            base_workflow=wf,
            dynamic_keys=(Streamed,),
            context_keys=(
                PrimaryContext,
                DerivedContext,
            ),  # DerivedContext depends on PrimaryContext
            target_keys=(Output,),
            accumulators=(Output,),
        )


def test_StreamProcessor_rejects_dynamic_keys_depending_on_context_keys() -> None:
    Context = NewType('Context', int)
    Streamed = NewType('Streamed', int)
    Output = NewType('Output', int)

    def derive_dynamic(context: Context) -> Streamed:
        return Streamed(context * 2)

    def process(value: Streamed) -> Output:
        return Output(value * 10)

    wf = sciline.Pipeline((derive_dynamic, process))

    # Dynamic key depending on context key should raise
    with pytest.raises(
        ValueError,
        match=r"Dynamic key .* depends on other dynamic/context keys: .*. "
        "This is not supported.",
    ):
        streaming.StreamProcessor(
            base_workflow=wf,
            dynamic_keys=(Streamed,),
            context_keys=(Context,),
            target_keys=(Output,),
            accumulators=(Output,),
        )


def test_StreamProcessor_rejects_context_keys_depending_on_dynamic_keys() -> None:
    Streamed = NewType('Streamed', int)
    Context = NewType('Context', int)
    Output = NewType('Output', int)

    def derive_context(value: Streamed) -> Context:
        return Context(value * 2)

    def process(value: Streamed, context: Context) -> Output:
        return Output(value + context)

    wf = sciline.Pipeline((derive_context, process))

    # Context key depending on dynamic key should raise
    with pytest.raises(
        ValueError,
        match=r"Context key .* depends on other dynamic/context keys: .*. "
        "This is not supported.",
    ):
        streaming.StreamProcessor(
            base_workflow=wf,
            dynamic_keys=(Streamed,),
            context_keys=(Context,),
            target_keys=(Output,),
            accumulators=(Output,),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

from ess.reduce import time_of_flight
from ess.reduce.time_of_flight import TofLookupTableWorkflow

sl = pytest.importorskip("sciline")


def test_lut_workflow_computes_table():
    wf = TofLookupTableWorkflow()
    wf[time_of_flight.DiskChoppers] = {}
    wf[time_of_flight.SourcePosition] = sc.vector([0, 0, 0], unit='m')
    wf[time_of_flight.NumberOfSimulatedNeutrons] = 100_000
    wf[time_of_flight.SimulationSeed] = 60
    wf[time_of_flight.PulseStride] = 1

    lmin, lmax = sc.scalar(25.0, unit='m'), sc.scalar(35.0, unit='m')
    dres = sc.scalar(0.1, unit='m')
    tres = sc.scalar(333.0, unit='us')

    wf[time_of_flight.LtotalRange] = lmin, lmax
    wf[time_of_flight.DistanceResolution] = dres
    wf[time_of_flight.TimeResolution] = tres

    table = wf.compute(time_of_flight.TimeOfFlightLookupTable)

    assert table.coords['distance'].min() < lmin
    assert table.coords['distance'].max() > lmax
    assert table.coords['event_time_offset'].max() == sc.scalar(1 / 14, unit='s').to(
        unit=table.coords['event_time_offset'].unit
    )
    assert sc.isclose(table.coords['distance_resolution'], dres)
    # Note that the time resolution is not exactly preserved since we want the table to
    # span exactly the frame period.
    assert sc.isclose(table.coords['time_resolution'], tres, rtol=sc.scalar(0.01))


def test_lut_workflow_computes_table_in_chunks():
    wf = TofLookupTableWorkflow()
    wf[time_of_flight.DiskChoppers] = {}
    wf[time_of_flight.SourcePosition] = sc.vector([0, 0, 0], unit='m')
    # Lots of neutrons activates chunking
    wf[time_of_flight.NumberOfSimulatedNeutrons] = 1_000_000
    wf[time_of_flight.SimulationSeed] = 61
    wf[time_of_flight.PulseStride] = 1

    lmin, lmax = sc.scalar(25.0, unit='m'), sc.scalar(35.0, unit='m')
    dres = sc.scalar(0.1, unit='m')
    tres = sc.scalar(250.0, unit='us')

    wf[time_of_flight.LtotalRange] = lmin, lmax
    wf[time_of_flight.DistanceResolution] = dres
    wf[time_of_flight.TimeResolution] = tres

    table = wf.compute(time_of_flight.TimeOfFlightLookupTable)

    assert table.coords['distance'].min() < lmin
    assert table.coords['distance'].max() > lmax
    assert table.coords['event_time_offset'].max() == sc.scalar(1 / 14, unit='s').to(
        unit=table.coords['event_time_offset'].unit
    )
    assert sc.isclose(table.coords['distance_resolution'], dres)
    # Note that the time resolution is not exactly preserved since we want the table to
    # span exactly the frame period.
    assert sc.isclose(table.coords['time_resolution'], tres, rtol=sc.scalar(0.01))


def test_lut_workflow_pulse_skipping():
    wf = TofLookupTableWorkflow()
    wf[time_of_flight.DiskChoppers] = {}
    wf[time_of_flight.SourcePosition] = sc.vector([0, 0, 0], unit='m')
    wf[time_of_flight.NumberOfSimulatedNeutrons] = 100_000
    wf[time_of_flight.SimulationSeed] = 62
    wf[time_of_flight.PulseStride] = 2

    lmin, lmax = sc.scalar(55.0, unit='m'), sc.scalar(65.0, unit='m')
    dres = sc.scalar(0.1, unit='m')
    tres = sc.scalar(250.0, unit='us')

    wf[time_of_flight.LtotalRange] = lmin, lmax
    wf[time_of_flight.DistanceResolution] = dres
    wf[time_of_flight.TimeResolution] = tres

    table = wf.compute(time_of_flight.TimeOfFlightLookupTable)

    assert table.coords['event_time_offset'].max() == 2 * sc.scalar(
        1 / 14, unit='s'
    ).to(unit=table.coords['event_time_offset'].unit)


def test_lut_workflow_non_exact_distance_range():
    wf = TofLookupTableWorkflow()
    wf[time_of_flight.DiskChoppers] = {}
    wf[time_of_flight.SourcePosition] = sc.vector([0, 0, 0], unit='m')
    wf[time_of_flight.NumberOfSimulatedNeutrons] = 100_000
    wf[time_of_flight.SimulationSeed] = 63
    wf[time_of_flight.PulseStride] = 1

    lmin, lmax = sc.scalar(25.0, unit='m'), sc.scalar(35.0, unit='m')
    dres = sc.scalar(0.33, unit='m')
    tres = sc.scalar(250.0, unit='us')

    wf[time_of_flight.LtotalRange] = lmin, lmax
    wf[time_of_flight.DistanceResolution] = dres
    wf[time_of_flight.TimeResolution] = tres

    table = wf.compute(time_of_flight.TimeOfFlightLookupTable)

    assert table.coords['distance'].min() < lmin
    assert table.coords['distance'].max() > lmax
    assert sc.isclose(table.coords['distance_resolution'], dres)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

from ess.reduce.live import roi


@pytest.fixture
def binned_indices() -> sc.DataArray:
    table = sc.data.table_xyz(123)
    table.coords['index'] = (
        10 * table.coords['x'] + 10 * table.coords['y'] + 10 * table.coords['z']
    ).to(dtype='int32')
    binned = table.bin(x=4, y=5)
    return sc.DataArray(binned.bins.coords['index'], coords=binned.coords)


def test_select_indices_positional_indexing(binned_indices):
    selected = roi.select_indices_in_intervals(
        intervals=sc.DataGroup(x=(1, 3), y=(2, 4)), indices=binned_indices
    )
    assert selected.dim == 'index'
    assert selected.sizes[selected.dim] > 0
    assert selected.sizes[selected.dim] < binned_indices.bins.size().sum().value


def test_select_indices_label_based_indexing(binned_indices):
    selected = roi.select_indices_in_intervals(
        intervals=sc.DataGroup(x=(sc.scalar(0.3, unit='m'), sc.scalar(0.5, unit='m'))),
        indices=binned_indices,
    )
    assert selected.dim == 'index'
    assert selected.sizes[selected.dim] > 0
    assert selected.sizes[selected.dim] < binned_indices.bins.size().sum().value


def test_select_indices_label_based_indexing_reverse_order(binned_indices):
    selected = roi.select_indices_in_intervals(
        intervals=sc.DataGroup(x=(sc.scalar(0.5, unit='m'), sc.scalar(0.3, unit='m'))),
        indices=binned_indices,
    )
    assert selected.dim == 'index'
    assert selected.sizes[selected.dim] > 0
    assert selected.sizes[selected.dim] < binned_indices.bins.size().sum().value


def test_select_indices_fails_with_invalid_dimension():
    data = sc.data.table_xyz(10)
    with pytest.raises(sc.DimensionError):
        roi.select_indices_in_intervals(
            intervals=sc.DataGroup(invalid_dim=(1, 2)), indices=data
        )


def test_select_indices_fails_without_required_coords():
    data = sc.DataArray(sc.array(dims=['x'], values=[1, 2, 3]))
    with pytest.raises(sc.DimensionError, match='no coordinate for that dimension'):
        roi.select_indices_in_intervals(
            intervals=sc.DataGroup(
                x=(sc.scalar(1.0, unit='m'), sc.scalar(2.0, unit='m'))
            ),
            indices=data,
        )


def test_select_indices_works_with_empty_selection(binned_indices):
    selected = roi.select_indices_in_intervals(
        intervals=sc.DataGroup(x=(1, 1)), indices=binned_indices
    )
    assert selected.dim == 'index'
    assert selected.sizes[selected.dim] == 0


def test_apply_selection_empty_yields_empty_result():
    selection = sc.array(dims=['index'], values=[], unit=None, dtype='int32')
    data = sc.arange('detector_number', 12, dtype='int32')
    result, _ = roi.apply_selection(data, selection=selection)
    assert sc.identical(
        result, sc.array(dims=['detector_number'], values=[], dtype='int32')
    )


def test_apply_selection_single_value():
    selection = sc.array(dims=['index'], values=[3], unit=None, dtype='int32')
    data = sc.arange('detector_number', 5, dtype='int32')
    result, _ = roi.apply_selection(data, selection=selection)
    assert sc.identical(
        result, sc.array(dims=['detector_number'], values=[3], dtype='int32')
    )


def test_apply_selection_repeated_indices():
    selection = sc.array(dims=['index'], values=[1, 1, 2, 1], unit=None, dtype='int32')
    data = sc.arange('detector_number', 5)
    result, scale = roi.apply_selection(data, selection=selection)
    assert sc.identical(result, sc.array(dims=['detector_number'], values=[1, 2]))
    assert sc.identical(scale, sc.array(dims=['detector_number'], values=[3.0, 1]))


def test_apply_selection_with_data_array():
    selection = sc.array(dims=['index'], values=[0, 0, 2], unit=None, dtype='int32')
    data = sc.DataArray(
        data=sc.arange('detector_number', 3, dtype='float64'),
        coords={'pos': sc.arange('detector_number', 3, unit='m', dtype='float64')},
    )
    result, _ = roi.apply_selection(data, selection=selection)
    expected = sc.DataArray(
        data=sc.array(dims=['detector_number'], values=[0.0, 2.0], dtype='float64')
        * sc.array(dims=['detector_number'], values=[2, 1], dtype='float64'),
        coords={'pos': sc.array(dims=['detector_number'], values=[0.0, 2.0], unit='m')},
    )
    assert sc.identical(result, expected)


def test_apply_selection_with_units():
    selection = sc.array(dims=['index'], values=[1, 1, 1], unit=None, dtype='int32')
    data = sc.arange('detector_number', 3, unit='K', dtype='float64')
    result, _ = roi.apply_selection(data, selection=selection)
    assert sc.identical(
        result, sc.array(dims=['detector_number'], values=[1.0], unit='K')
    )


def test_apply_selection_non_contiguous_indices():
    selection = sc.array(dims=['index'], values=[0, 4, 4, 8], unit=None, dtype='int32')
    data = sc.arange('detector_number', 10)
    result, scale = roi.apply_selection(data, selection=selection)
    assert sc.identical(result, sc.array(dims=['detector_number'], values=[0, 4, 8]))
    assert sc.identical(scale, sc.array(dims=['detector_number'], values=[1.0, 2, 1]))


def test_apply_selection_ignores_selection_dim():
    selection = sc.array(dims=['ignored'], values=[1, 2], dtype='int32')
    data = sc.arange('detector_number', 5, dtype='int32')
    result, _ = roi.apply_selection(data, selection=selection)
    assert sc.identical(
        result, sc.array(dims=['detector_number'], values=[1, 2], dtype='int32')
    )


def test_apply_selection_fails_with_out_of_bounds_index():
    selection = sc.array(dims=['index'], values=[10], dtype='int32')
    data = sc.arange('detector_number', 5, dtype='int32')
    with pytest.raises(IndexError):
        roi.apply_selection(data, selection=selection)


def logical_view(da: sc.DataArray) -> sc.DataArray:
    return da.fold(da.dim, sizes={'x': 3, 'y': 4, 'z': 2})['z', 0]


@pytest.fixture
def roi_filter() -> roi.ROIFilter:
    indices = sc.ones(sizes={'detector_number': 24}, dtype='int32', unit=None)
    indices = sc.cumsum(indices, mode='exclusive')
    return roi.ROIFilter(logical_view(indices))


def test_ROIFilter_defaults_to_empty_roi(roi_filter: roi.ROIFilter):
    data = sc.linspace('detector_number', 0.0, 1.0, num=24, unit='counts')
    result, _ = roi_filter.apply(data)
    assert sc.identical(
        result, sc.array(dims=['detector_number'], values=[], unit='counts')
    )


def test_ROIFilter_applies_roi(roi_filter: roi.ROIFilter):
    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(1, 3), y=(2, 4)))
    data = sc.linspace('detector_number', 1.0, 24.0, num=24, unit='counts')
    result, _ = roi_filter.apply(data)
    assert sc.identical(
        result,
        sc.array(
            dims=['detector_number'], values=[13.0, 15.0, 21.0, 23.0], unit='counts'
        ),
    )


def test_ROIFilter_applies_roi_to_2d_data(roi_filter: roi.ROIFilter):
    roi_filter.set_roi_from_intervals(sc.DataGroup(x=(1, 3), y=(2, 4)))
    data = sc.linspace('detector_number', 1.0, 24.0, num=24, unit='counts').fold(
        dim='detector_number', sizes={'x': 3, 'y': 4, 'z': 2}
    )
    result, _ = roi_filter.apply(data)
    assert sc.identical(
        result,
        sc.array(
            dims=['detector_number'], values=[13.0, 15.0, 21.0, 23.0], unit='counts'
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""Utilities for region of interest (ROI) selection."""

from __future__ import annotations

from typing import TypeVar

import numpy as np
import scipp as sc


def select_indices_in_intervals(
    intervals: sc.DataGroup[tuple[int, int] | tuple[sc.Variable, sc.Variable]],
    indices: sc.Variable | sc.DataArray,
) -> sc.Variable:
    """
    Return subset of indices that fall within the intervals.

    Parameters
    ----------
    intervals:
        DataGroup with dimension names as keys and tuples of low and high values. This
        can be used to define a band or a rectangle to selected. When low and high are
        scipp.Variable, the selection is done using label-based indexing. In this case
        `indices` must be a DataArray with corresponding coordinates.
    indices:
        Variable or DataArray with indices to select from. If binned data the selected
        indices will be returned concatenated into a dense array.
    """
    out_dim = 'index'
    for dim, bounds in intervals.items():
        low, high = sorted(bounds)
        indices = indices[dim, low:high]
    indices = indices if isinstance(indices, sc.Variable) else indices.data
    indices = indices.flatten(to=out_dim)
    if indices.bins is None:
        return indices
    indices = indices.bins.concat().value
    return indices.rename_dims({indices.dim: out_dim})


T = TypeVar('T', sc.DataArray, sc.Variable)


def apply_selection(
    data: T, *, selection: sc.Variable, norm: float = 1.0
) -> tuple[T, sc.Variable]:
    """
    Apply selection to data.

    Parameters
    ----------
    data:
        Data to filter.
    selection:
        Variable with indices to select.
    norm:
        Normalization factor to apply to the selected data. This is used for cases where
        indices may be selected multiple times.

    Returns
    -------
    :
        Filtered data and scale factor.
    """
    indices, counts = np.unique(selection.values, return_counts=True)
    if data.ndim != 1:
        data = data.flatten(to='detector_number')
    scale = sc.array(dims=[data.dim], values=counts) / norm
    return data[indices], scale


class ROIFilter:
    """Filter for selecting a region of interest (ROI)."""

    def __init__(self, indices: sc.Variable | sc.DataArray, norm: float = 1.0) -> None:
        """
        Create a new ROI filter.

        Parameters
        ----------
        indices:
            Variable with indices to filter. The indices facilitate selecting a 2-D
            ROI in a projection of a 3-D dataset. Typically the indices are given by a
            2-D array. Each element in the array may correspond to a single index (when
            there is no projection) or a list of indices that were projected into an
            output pixel.
        """
        self._indices = indices
        self._selection = sc.array(dims=['index'], values=[])
        self._norm = norm

    def set_roi_from_intervals(self, intervals: sc.DataGroup) -> None:
        """Set the ROI from (typically 1 or 2) intervals."""
        self._selection = select_indices_in_intervals(intervals, self._indices)

    def apply(self, data: T) -> tuple[T, sc.Variable]:
        """
        Apply the ROI filter to data.

        The returned scale factor can be used to handle filtering via a projection, to
        take into account that fractions of source data point contribute to a data point
        in the projection.

        Parameters
        ----------
        data:
            Data to filter.

        Returns
        -------
        :
            Filtered data and scale factor.
        """
        return apply_selection(data, selection=self._selection, norm=self._norm)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc
from scippneutron.chopper import DiskChopper
from scippneutron.conversion.graph.beamline import beamline as beamline_graph
from scippneutron.conversion.graph.tof import elastic as elastic_graph

from ess.reduce import time_of_flight
from ess.reduce.nexus.types import DetectorData, SampleRun
from ess.reduce.time_of_flight import GenericTofWorkflow, TofLookupTableWorkflow, fakes

sl = pytest.importorskip("sciline")


def make_lut_workflow(choppers, neutrons, seed, pulse_stride):
    lut_wf = TofLookupTableWorkflow()
    lut_wf[time_of_flight.DiskChoppers] = choppers
    lut_wf[time_of_flight.SourcePosition] = fakes.source_position()
    lut_wf[time_of_flight.NumberOfSimulatedNeutrons] = neutrons
    lut_wf[time_of_flight.SimulationSeed] = seed
    lut_wf[time_of_flight.PulseStride] = pulse_stride
    lut_wf[time_of_flight.SimulationResults] = lut_wf.compute(
        time_of_flight.SimulationResults
    )
    return lut_wf


@pytest.fixture(scope="module")
def lut_workflow_psc_choppers():
    return make_lut_workflow(
        choppers=fakes.psc_choppers(), neutrons=500_000, seed=1234, pulse_stride=1
    )


@pytest.fixture(scope="module")
def lut_workflow_pulse_skipping():
    return make_lut_workflow(
        choppers=fakes.pulse_skipping_choppers(),
        neutrons=500_000,
        seed=112,
        pulse_stride=2,
    )


def _make_workflow_event_mode(
    distance,
    choppers,
    lut_workflow,
    seed,
    pulse_stride_offset,
    error_threshold,
):
    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=300_000,
        seed=seed,
    )
    mon, ref = beamline.get_monitor("detector")

    pl = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])
    pl[DetectorData[SampleRun]] = mon
    pl[time_of_flight.DetectorLtotal[SampleRun]] = distance
    pl[time_of_flight.PulseStrideOffset] = pulse_stride_offset

    lut_wf = lut_workflow.copy()
    lut_wf[time_of_flight.LtotalRange] = distance, distance
    lut_wf[time_of_flight.LookupTableRelativeErrorThreshold] = error_threshold

    pl[time_of_flight.TimeOfFlightLookupTable] = lut_wf.compute(
        time_of_flight.TimeOfFlightLookupTable
    )

    return pl, ref


def _make_workflow_histogram_mode(dim, distance, choppers, lut_workflow, seed):
    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=100_000,
        seed=seed,
    )
    mon, ref = beamline.get_monitor("detector")
    mon = mon.hist(
        event_time_offset=sc.linspace(
            "event_time_offset", 0.0, 1000.0 / 14, num=301, unit="ms"
        ).to(unit=mon.bins.coords["event_time_offset"].bins.unit)
    ).rename(event_time_offset=dim)

    pl = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])
    pl[DetectorData[SampleRun]] = mon
    pl[time_of_flight.DetectorLtotal[SampleRun]] = distance

    lut_wf = lut_workflow.copy()
    lut_wf[time_of_flight.LtotalRange] = distance, distance

    pl[time_of_flight.TimeOfFlightLookupTable] = lut_wf.compute(
        time_of_flight.TimeOfFlightLookupTable
    )

    return pl, ref


def _validate_result_events(tofs, ref, percentile, diff_threshold, rtol):
    assert "event_time_offset" not in tofs.coords

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # Most errors should be small
    assert np.nanpercentile(diff.values, percentile) < diff_threshold
    # Make sure that we have not lost too many events (we lose some because they may be
    # given a NaN tof from the lookup).
    nevents = sc.sum(~sc.isnan(wavs.coords['wavelength'])).to(dtype=float)
    nevents.unit = 'counts'
    assert sc.isclose(ref.data.sum(), nevents, rtol=sc.scalar(rtol))


def _validate_result_histogram_mode(tofs, ref, percentile, diff_threshold, rtol):
    assert "time_of_flight" not in tofs.coords
    assert "frame_time" not in tofs.coords

    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph)
    ref = ref.hist(wavelength=wavs.coords["wavelength"])
    # We divide by the maximum to avoid large relative differences at the edges of the
    # frames where the counts are low.
    diff = (wavs - ref) / ref.max()
    assert np.nanpercentile(diff.values, percentile) < diff_threshold
    # Make sure that we have not lost too many events (we lose some because they may be
    # given a NaN tof from the lookup).
    assert sc.isclose(ref.data.nansum(), tofs.data.nansum(), rtol=sc.scalar(rtol))


def test_unwrap_with_no_choppers() -> None:
    # At this small distance the frames are not overlapping (with the given wavelength
    # range), despite not using any choppers.
    distance = sc.scalar(10.0, unit="m")
    choppers = {}

    lut_wf = make_lut_workflow(
        choppers=choppers, neutrons=300_000, seed=1234, pulse_stride=1
    )

    pl, ref = _make_workflow_event_mode(
        distance=distance,
        choppers=choppers,
        lut_workflow=lut_wf,
        seed=1,
        pulse_stride_offset=0,
        error_threshold=1.0,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=96, diff_threshold=1.0, rtol=0.02
    )


# At 30m, event_time_offset does not wrap around (all events within the first pulse).
# At 60m, all events are within the second pulse.
# At 80m, events are split between the second and third pulse.
# At 108m, events are split between the third and fourth pulse.
@pytest.mark.parametrize("dist", [30.0, 60.0, 80.0, 108.0])
def test_standard_unwrap(dist, lut_workflow_psc_choppers) -> None:
    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(dist, unit="m"),
        choppers=fakes.psc_choppers(),
        lut_workflow=lut_workflow_psc_choppers,
        seed=2,
        # pulse_stride=1,
        pulse_stride_offset=0,
        error_threshold=0.1,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.02, rtol=0.05
    )


# At 30m, event_time_offset does not wrap around (all events within the first pulse).
# At 60m, all events are within the second pulse.
# At 80m, events are split between the second and third pulse.
# At 108m, events are split between the third and fourth pulse.
@pytest.mark.parametrize("dist", [30.0, 60.0, 80.0, 108.0])
@pytest.mark.parametrize("dim", ["time_of_flight", "tof", "frame_time"])
def test_standard_unwrap_histogram_mode(dist, dim, lut_workflow_psc_choppers) -> None:
    pl, ref = _make_workflow_histogram_mode(
        dim=dim,
        distance=sc.scalar(dist, unit="m"),
        choppers=fakes.psc_choppers(),
        lut_workflow=lut_workflow_psc_choppers,
        seed=37,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_histogram_mode(
        tofs=tofs, ref=ref, percentile=96, diff_threshold=0.4, rtol=0.05
    )


@pytest.mark.parametrize("dist", [60.0, 100.0])
def test_pulse_skipping_unwrap(dist, lut_workflow_pulse_skipping) -> None:
    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(dist, unit="m"),
        choppers=fakes.pulse_skipping_choppers(),
        lut_workflow=lut_workflow_pulse_skipping,
        seed=432,
        # pulse_stride=2,
        pulse_stride_offset=1,
        error_threshold=0.1,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.1, rtol=0.05
    )


def test_pulse_skipping_unwrap_180_phase_shift() -> None:
    choppers = fakes.pulse_skipping_choppers()
    choppers["pulse_skipping"].phase.value += 180.0

    lut_wf = make_lut_workflow(
        choppers=choppers, neutrons=500_000, seed=111, pulse_stride=2
    )

    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(100.0, unit="m"),
        choppers=choppers,
        lut_workflow=lut_wf,
        seed=55,
        pulse_stride_offset=1,
        error_threshold=0.1,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.1, rtol=0.05
    )


@pytest.mark.parametrize("dist", [60.0, 100.0])
def test_pulse_skipping_stride_offset_guess_gives_expected_result(
    dist, lut_workflow_pulse_skipping
) -> None:
    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(dist, unit="m"),
        choppers=fakes.pulse_skipping_choppers(),
        lut_workflow=lut_workflow_pulse_skipping,
        seed=97,
        pulse_stride_offset=None,
        error_threshold=0.1,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.1, rtol=0.05
    )


def test_pulse_skipping_unwrap_when_all_neutrons_arrive_after_second_pulse() -> None:
    choppers = fakes.pulse_skipping_choppers()
    choppers['chopper'] = DiskChopper(
        frequency=sc.scalar(-14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-35.0, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 8.0], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=np.array([10.0]), unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=np.array([25.0]), unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    lut_wf = make_lut_workflow(
        choppers=choppers, neutrons=500_000, seed=222, pulse_stride=2
    )

    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(150.0, unit="m"),
        choppers=choppers,
        lut_workflow=lut_wf,
        seed=6,
        pulse_stride_offset=1,
        error_threshold=0.1,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.1, rtol=0.05
    )


def test_pulse_skipping_unwrap_when_first_half_of_first_pulse_is_missing() -> None:
    distance = sc.scalar(100.0, unit="m")
    choppers = fakes.pulse_skipping_choppers()

    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1.0, unit="s"),
        events_per_pulse=100_000,
        seed=21,
    )
    mon, ref = beamline.get_monitor("detector")

    lut_wf = make_lut_workflow(
        choppers=choppers, neutrons=300_000, seed=1234, pulse_stride=2
    )
    lut_wf[time_of_flight.LtotalRange] = distance, distance

    pl = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])

    # Skip first pulse = half of the first frame
    a = mon.group('event_time_zero')['event_time_zero', 1:]
    a.bins.coords['event_time_zero'] = sc.bins_like(a, a.coords['event_time_zero'])
    pl[DetectorData[SampleRun]] = a.bins.concat('event_time_zero')
    pl[time_of_flight.DetectorLtotal[SampleRun]] = distance

    pl[time_of_flight.TimeOfFlightLookupTable] = lut_wf.compute(
        time_of_flight.TimeOfFlightLookupTable
    )
    pl[time_of_flight.PulseStrideOffset] = 1  # Start the stride at the second pulse

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value
    # Bin the events in toa starting from the pulse period to skip the first pulse.
    ref = (
        ref.bin(
            toa=sc.concat(
                [
                    sc.scalar(1 / 14, unit='s').to(unit=ref.coords['toa'].unit),
                    ref.coords['toa'].max() * 1.01,
                ],
                dim='toa',
            )
        )
        .bins.concat()
        .value
    )

    # Sort the events according id to make sure we are comparing the same values.
    wavs = sc.sort(wavs, key=wavs.coords['id'])
    ref = sc.sort(ref, key=ref.coords['id'])

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # All errors should be small
    assert np.nanpercentile(diff.values, 100) < 0.05
    # Make sure that we have not lost too many events (we lose some because they may be
    # given a NaN tof from the lookup).
    assert sc.isclose(
        pl.compute(DetectorData[SampleRun]).data.nansum(),
        tofs.data.nansum(),
        rtol=sc.scalar(1.0e-3),
    )


def test_pulse_skipping_stride_3() -> None:
    choppers = fakes.pulse_skipping_choppers()
    choppers["pulse_skipping"].frequency.value = -14.0 / 3.0

    lut_wf = make_lut_workflow(
        choppers=choppers, neutrons=500_000, seed=111, pulse_stride=3
    )

    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(150.0, unit="m"),
        choppers=choppers,
        lut_workflow=lut_wf,
        seed=68,
        pulse_stride_offset=None,
        error_threshold=0.1,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.1, rtol=0.05
    )


def test_pulse_skipping_unwrap_histogram_mode(lut_workflow_pulse_skipping) -> None:
    pl, ref = _make_workflow_histogram_mode(
        dim='time_of_flight',
        distance=sc.scalar(50.0, unit="m"),
        choppers=fakes.pulse_skipping_choppers(),
        lut_workflow=lut_workflow_pulse_skipping,
        seed=9,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_histogram_mode(
        tofs=tofs, ref=ref, percentile=96, diff_threshold=0.4, rtol=0.05
    )


@pytest.mark.parametrize("dtype", ["int32", "int64"])
def test_unwrap_int(dtype, lut_workflow_psc_choppers) -> None:
    pl, ref = _make_workflow_event_mode(
        distance=sc.scalar(80.0, unit="m"),
        choppers=fakes.psc_choppers(),
        lut_workflow=lut_workflow_psc_choppers,
        seed=2,
        pulse_stride_offset=0,
        error_threshold=0.1,
    )

    mon = pl.compute(DetectorData[SampleRun]).copy()
    mon.bins.coords["event_time_offset"] = mon.bins.coords["event_time_offset"].to(
        dtype=dtype, unit="ns"
    )
    pl[DetectorData[SampleRun]] = mon

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    _validate_result_events(
        tofs=tofs, ref=ref, percentile=100, diff_threshold=0.02, rtol=0.05
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
A fake time-of-flight neutron beamline for documentation and testing.

This provides detector event data in a structure as typically provided in a NeXus file,
with event_time_offset and event_time_zero information.
"""

from collections.abc import Callable

import numpy as np
import scipp as sc
from scippneutron.chopper import DiskChopper


class FakeBeamline:
    def __init__(
        self,
        choppers: dict[str, DiskChopper],
        monitors: dict[str, sc.Variable],
        run_length: sc.Variable,
        events_per_pulse: int = 200000,
        seed: int | None = None,
        source: Callable | None = None,
        source_position: sc.Variable | None = None,
    ):
        import math

        import tof as tof_pkg
        from tof.facilities.ess_pulse import frequency as ess_frequency

        self.frequency = ess_frequency
        self.npulses = math.ceil((run_length * self.frequency).to(unit="").value)
        self.events_per_pulse = events_per_pulse
        if source_position is None:
            source_position = sc.vector([0, 0, 0], unit='m')

        # Create a source
        if source is None:
            self.source = tof_pkg.Source(
                facility="ess",
                neutrons=self.events_per_pulse,
                pulses=self.npulses,
                seed=seed,
            )
        else:
            self.source = source(pulses=self.npulses)

        # Convert the choppers to tof.Chopper
        self.choppers = [
            tof_pkg.Chopper(
                frequency=abs(ch.frequency),
                direction=tof_pkg.AntiClockwise
                if (ch.frequency.value > 0.0)
                else tof_pkg.Clockwise,
                open=ch.slit_begin,
                close=ch.slit_end,
                phase=abs(ch.phase),
                distance=sc.norm(ch.axle_position - source_position),
                name=name,
            )
            for name, ch in choppers.items()
        ]

        # Add detectors
        self.monitors = [
            tof_pkg.Detector(distance=distance, name=key)
            for key, distance in monitors.items()
        ]

        #  Propagate the neutrons
        self.model = tof_pkg.Model(
            source=self.source, choppers=self.choppers, detectors=self.monitors
        )
        self.model_result = self.model.run()

    def get_monitor(self, name: str) -> sc.DataGroup:
        nx_event_data = self.model_result.to_nxevent_data(name)
        raw_data = self.model_result.detectors[name].data.flatten(to="event")
        raw_data = raw_data[~raw_data.masks["blocked_by_others"]].copy()
        return nx_event_data, raw_data


def psc_choppers():
    return {
        "chopper": DiskChopper(
            frequency=sc.scalar(-14.0, unit="Hz"),
            beam_position=sc.scalar(0.0, unit="deg"),
            phase=sc.scalar(-85.0, unit="deg"),
            axle_position=sc.vector(value=[0, 0, 8.0], unit="m"),
            slit_begin=sc.array(dims=["cutout"], values=[0.0], unit="deg"),
            slit_end=sc.array(dims=["cutout"], values=[3.0], unit="deg"),
            slit_height=sc.scalar(10.0, unit="cm"),
            radius=sc.scalar(30.0, unit="cm"),
        )
    }


def pulse_skipping_choppers():
    return {
        "chopper": DiskChopper(
            frequency=sc.scalar(-14.0, unit="Hz"),
            beam_position=sc.scalar(0.0, unit="deg"),
            phase=sc.scalar(-35.0, unit="deg"),
            axle_position=sc.vector(value=[0, 0, 8.0], unit="m"),
            slit_begin=sc.array(dims=["cutout"], values=np.array([0.0]), unit="deg"),
            slit_end=sc.array(dims=["cutout"], values=np.array([33.0]), unit="deg"),
            slit_height=sc.scalar(10.0, unit="cm"),
            radius=sc.scalar(30.0, unit="cm"),
        ),
        "pulse_skipping": DiskChopper(
            frequency=sc.scalar(-7.0, unit="Hz"),
            beam_position=sc.scalar(0.0, unit="deg"),
            phase=sc.scalar(-10.0, unit="deg"),
            axle_position=sc.vector(value=[0, 0, 15.0], unit="m"),
            slit_begin=sc.array(dims=["cutout"], values=np.array([0.0]), unit="deg"),
            slit_end=sc.array(dims=["cutout"], values=np.array([120.0]), unit="deg"),
            slit_height=sc.scalar(10.0, unit="cm"),
            radius=sc.scalar(30.0, unit="cm"),
        ),
    }


def source_position():
    return sc.vector([0, 0, 0], unit='m')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import numpy as np


class Interpolator:
    def __init__(
        self,
        time_edges: np.ndarray,
        distance_edges: np.ndarray,
        values: np.ndarray,
        method: str = "linear",
        bounds_error: bool = False,
        fill_value: float = np.nan,
        **kwargs,
    ):
        """
        Interpolator for 2D regular grid data (SciPy implementation).

        Parameters
        ----------
        time_edges:
            1D array of time edges (length N_time).
        distance_edges:
            1D array of distance edges (length N_dist).
        values:
            2D array of values on the grid. The shape must be (N_dist, N_time).
        method:
            Method of interpolation. Default is "linear".
        bounds_error:
            If True, when interpolated values are requested outside of the domain,
            a ValueError is raised. If False, fill_value is used.
        fill_value:
            Value to use for points outside of the grid.
        kwargs:
            Additional arguments to pass to scipy.interpolate.RegularGridInterpolator.
        """
        from scipy.interpolate import RegularGridInterpolator

        self._interp = RegularGridInterpolator(
            (distance_edges, time_edges),
            values,
            method=method,
            bounds_error=bounds_error,
            fill_value=fill_value,
            **kwargs,
        )

    def __call__(
        self,
        times: np.ndarray,
        distances: np.ndarray,
        pulse_period: float = 0.0,
        pulse_index: np.ndarray | None = None,
    ) -> np.ndarray:
        if pulse_index is not None:
            times = times + (pulse_index * pulse_period)
        return self._interp((distances, times))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""Data files bundled with ESSreduce.

This module requires the Pooch package which is not a hard dependency of ESSreduce.
It has to be installed separately with either pip or conda.
"""

from ._registry import Entry, LocalRegistry, PoochRegistry, Registry, make_registry

__all__ = [
    'Entry',
    'LocalRegistry',
    'PoochRegistry',
    'Registry',
    'make_registry',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""Logging tools for ess.reduce."""

import logging


def get_logger() -> logging.Logger:
    """Return the logger for ess.reduce.

    Returns
    -------
    :
        The requested logger.
    """
    return logging.getLogger('scipp.ess.reduce')
